"""
An implementation of the SMV image reader for Rigaku Eiger.
Be aware: this is completely unrelated to the HDF5 Eiger format.
"""

from __future__ import annotations

import sys
import time

from boost_adaptbx.boost.python import streambuf
from cctbx.eltbx import attenuation_coefficient
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.ext import read_int32
from dxtbx.format.FormatSMVRigaku import FormatSMVRigaku
from dxtbx.model import ParallaxCorrectedPxMmStrategy


class FormatSMVRigakuEiger(FormatSMVRigaku):
    """A class for reading SMV format Rigaku Eiger images."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like a Rigaku d*TREK SMVRigaku format image,
        i.e. we can make sense of it. Essentially that will be if it contains
        all of the keys we are looking for."""

        size, header = FormatSMVRigaku.get_smv_header(image_file)

        if "DETECTOR_NAMES" not in header:
            return False

        name = header["DETECTOR_NAMES"]

        if "%sDETECTOR_DESCRIPTION" % name not in header:
            return False

        return "Eiger" in header["%sDETECTOR_DESCRIPTION" % name]

    def _goniometer(self):
        """Initialize the structure for the goniometer."""

        values = [
            float(e) for e in self._header_dictionary["CRYSTAL_GONIO_VALUES"].split()
        ]
        names = [
            e.strip() for e in self._header_dictionary["CRYSTAL_GONIO_NAMES"].split()
        ]
        units = [
            e.strip() for e in self._header_dictionary["CRYSTAL_GONIO_UNITS"].split()
        ]
        axis_elts = [
            float(e) for e in self._header_dictionary["CRYSTAL_GONIO_VECTORS"].split()
        ]
        axes = [matrix.col(axis_elts[3 * j : 3 * (j + 1)]) for j in range(len(units))]
        scan_axis = self._header_dictionary["ROTATION_AXIS_NAME"].strip()

        # Take only elements that have corresponding units of 'deg' (which is
        # probably all of them).
        filt = [e == "deg" for e in units]
        values = [e for e, f in zip(values, filt) if f]
        names = [e for e, f in zip(names, filt) if f]
        axes = [e for e, f in zip(axes, filt) if f]

        # Multi-axis gonio requires axes in order as viewed from crystal to gonio
        # base. Assume the SMV header records them in reverse order.

        axes = flex.vec3_double(reversed(axes))
        names = flex.std_string(reversed(names))
        values = flex.double(reversed(values))
        scan_axis = flex.first_index(names, scan_axis)

        gonio = self._goniometer_factory.make_multi_axis_goniometer(
            axes, values, names, scan_axis
        )

        # The calculated rotation axis is also recorded in the header. We could
        # use this to check that the goniometer is as expected
        rot_axis = tuple(map(float, self._header_dictionary["ROTATION_VECTOR"].split()))
        for e1, e2 in zip(rot_axis, gonio.get_rotation_axis()):
            assert abs(e1 - e2) < 1e-6

        return gonio

    def _detector(self):
        """Return a model for the detector, allowing for two-theta offsets
        and the detector position. This will be rather more complex..."""

        detector_name = self._header_dictionary["DETECTOR_NAMES"].split()[0].strip()

        detector_axes = self.get_detector_axes(detector_name)
        fast = matrix.col(tuple(detector_axes[:3]))
        slow = matrix.col(tuple(detector_axes[3:]))
        distortion = self.get_distortion(detector_name)

        # multiply through by the distortion to get the true detector fast, slow

        detector_fast, detector_slow = (
            distortion[0] * fast + distortion[1] * slow,
            distortion[2] * fast + distortion[3] * slow,
        )

        beam_pixels = self.get_beam_pixels(detector_name)
        pixel_size = self.get_pixel_size(detector_name)
        image_size = self.get_image_size(detector_name)

        detector_origin = -(
            beam_pixels[0] * pixel_size[0] * detector_fast
            + beam_pixels[1] * pixel_size[1] * detector_slow
        )

        gonio_axes = self.get_gonio_axes(detector_name)
        gonio_values = self.get_gonio_values(detector_name)
        gonio_units = self._header_dictionary["%sGONIO_UNITS" % detector_name].split()
        gonio_num_axes = int(
            self._header_dictionary["%sGONIO_NUM_VALUES" % detector_name]
        )

        rotations = []
        translations = []

        for j, unit in enumerate(gonio_units):
            axis = matrix.col(gonio_axes[3 * j : 3 * (j + 1)])
            if unit == "deg":
                rotations.append(
                    axis.axis_and_angle_as_r3_rotation_matrix(gonio_values[j], deg=True)
                )
                translations.append(matrix.col((0.0, 0.0, 0.0)))
            elif unit == "mm":
                rotations.append(
                    matrix.sqr((1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0))
                )
                translations.append(gonio_values[j] * axis)
            else:
                raise RuntimeError("unknown axis unit %s" % unit)

        rotations.reverse()
        translations.reverse()

        for j in range(gonio_num_axes):
            detector_fast = rotations[j] * detector_fast
            detector_slow = rotations[j] * detector_slow
            detector_origin = rotations[j] * detector_origin
            detector_origin = translations[j] + detector_origin

        max_trusted_value = int(float(self._header_dictionary["SATURATED_VALUE"]))
        min_trusted_value = 0

        # Unfortunately thickness and material are not stored in the header. Set
        # to sensible defaults.
        t0 = 0.450
        material = "Si"

        table = attenuation_coefficient.get_table(material)
        wavelength = float(self._header_dictionary["SCAN_WAVELENGTH"])
        mu = table.mu_at_angstrom(wavelength) / 10.0

        detector = self._detector_factory.complex(
            "PAD",
            detector_origin.elems,
            detector_fast.elems,
            detector_slow.elems,
            pixel_size,
            image_size,
            (min_trusted_value, max_trusted_value),
            px_mm=ParallaxCorrectedPxMmStrategy(mu, t0),
            mu=mu,
        )

        for panel in detector:
            panel.set_thickness(t0)
            panel.set_material(material)

        return detector

    def _beam(self):
        """Return a simple model for the beam."""

        beam_direction = self.get_beam_direction()
        p_fraction, p_plane = self.get_beam_polarization()

        wavelength = float(self._header_dictionary["SCAN_WAVELENGTH"])

        return self._beam_factory.complex(
            beam_direction, p_fraction, p_plane, wavelength
        )

    def _scan(self):
        """Return the scan information for this image."""
        epoch = time.strptime(
            self._header_dictionary["DTREK_DATE_TIME"], "%d-%b-%Y %H:%M:%S"
        )
        return self._create_single_SVM_scan(epoch)

    def get_raw_data(self):
        """Read the data - assuming it is streaming 4-byte unsigned ints following the
        header..."""

        assert len(self.get_detector()) == 1
        size = self.get_detector()[0].get_image_size()
        f = self.open_file(self._image_file)
        f.read(int(self._header_dictionary["HEADER_BYTES"]))
        raw_data = read_int32(streambuf(f), int(size[0] * size[1]))
        raw_data.reshape(flex.grid(size[1], size[0]))

        return raw_data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVRigakuEiger.understand(arg))



from __future__ import annotations

import math

import pycbf

import libtbx.phil
from scitbx.array_family import flex

try:
    from ..dxtbx_model_ext import Goniometer, KappaGoniometer, MultiAxisGoniometer
except ModuleNotFoundError:
    from dxtbx_model_ext import (  # type: ignore
        Goniometer,
        KappaGoniometer,
        MultiAxisGoniometer,
    )

__all__ = [
    "Goniometer",
    "GoniometerFactory",
    "KappaGoniometer",
    "MultiAxisGoniometer",
    "goniometer_phil_scope",
]

goniometer_phil_scope = libtbx.phil.parse(
    """
  goniometer
    .expert_level = 1
    .short_caption = "Goniometer overrides"
  {

    axis = None
      .type = floats(size=3)
      .help = "Override the axis for a single axis goniometer. Equivalent to"
              "providing a single 3D vector to 'axes'."
      .short_caption="Goniometer axis"

    axes = None
      .type = floats
      .help = "Override the goniometer axes. Axes must be provided in the"
              "order crystal-to-goniometer, i.e. for a Kappa goniometer"
              "phi,kappa,omega"
      .short_caption="Goniometer axes"

    angles = None
      .type = floats
      .help = "Override the goniometer angles. Axes must be provided in the"
              "order crystal-to-goniometer, i.e. for a Kappa goniometer"
              "phi,kappa,omega"
      .short_caption = "Goniometer angles"

    names = None
      .type = str
      .help = "The multi axis goniometer axis names"
      .short_caption = "The axis names"

    scan_axis = None
      .type = int
      .help = "The scan axis"
      .short_caption = "The scan axis"

    fixed_rotation = None
      .type = floats(size=9)
      .help = "Override the fixed rotation matrix"
      .short_caption = "Fixed rotation matrix"

    setting_rotation = None
      .type = floats(size=9)
      .help = "Override the setting rotation matrix"
      .short_caption = "Setting rotation matrix"

    invert_rotation_axis = False
      .type = bool
      .help = "Invert the rotation axis"
      .short_caption = "Invert rotation axis"

  }
"""
)


class GoniometerFactory:
    """A factory class for goniometer objects, which will encapsulate
    some standard goniometer designs to make it a little easier to get
    started with all of this - for cases when we are not using a CBF.
    When we have a CBF just use that factory method and everything will be
    peachy."""

    @staticmethod
    def single_axis_goniometer_from_phil(params, reference=None):
        """
        Generate or overwrite a single axis goniometer

        """

        # Check the axis parameter and copy from axes if required
        if params.goniometer.axis is None:
            params.goniometer.axis = params.goniometer.axes
        if params.goniometer.axis is not None and len(params.goniometer.axis) != 3:
            raise RuntimeError("Single axis goniometer requires 3 axis values")

        # Check the angles parameter
        if params.goniometer.angles is not None:
            raise RuntimeError("Single axis goniometer requires angles == None")

        # Check the names parameter
        if params.goniometer.names is not None:
            raise RuntimeError("Single axis goniometer requires names == None")

        # Init the gonionmeter
        if reference is None:
            goniometer = Goniometer()
        else:
            goniometer = reference

        # Set the parameters
        if params.goniometer.axis is not None:
            goniometer.set_rotation_axis_datum(params.goniometer.axis)
        if params.goniometer.fixed_rotation is not None:
            goniometer.set_fixed_rotation(params.goniometer.fixed_rotation)
        if params.goniometer.setting_rotation is not None:
            goniometer.set_setting_rotation(params.goniometer.setting_rotation)
        if params.goniometer.invert_rotation_axis is True:
            rotation_axis = goniometer.get_rotation_axis_datum()
            goniometer.set_rotation_axis_datum([-x for x in rotation_axis])

        return goniometer

    @staticmethod
    def multi_axis_goniometer_from_phil(params, reference=None):
        # Check the axes parameter
        if params.goniometer.axes is not None:
            if len(params.goniometer.axes) % 3:
                raise RuntimeError(
                    "Number of values for axes parameter must be multiple of 3."
                )

        # Check the fixed rotation
        if params.goniometer.fixed_rotation is not None:
            raise RuntimeError("Multi-axis goniometer requires fixed_rotation == None")

        # Check the setting rotation
        if params.goniometer.setting_rotation is not None:
            raise RuntimeError(
                "Multi-axis goniometer requires setting_rotation == None"
            )

        # Check the input
        if reference is None:
            if params.goniometer.axes is None:
                raise RuntimeError("No axes set")

            # Create the axes
            axes = flex.vec3_double(
                params.goniometer.axes[i * 3 : (i * 3) + 3]
                for i in range(len(params.goniometer.axes) // 3)
            )

            # Invert the rotation axis
            if params.goniometer.invert_rotation_axis is True:
                axes = flex.vec3_double([[-x for x in v] for v in axes])

            # Create the angles
            if params.goniometer.angles is not None:
                angles = params.goniometer.angles
                if len(angles) != len(axes):
                    raise RuntimeError("Number of angles must match axes")
            else:
                angles = flex.double([0] * len(axes))

            # Create the names
            if params.goniometer.names is not None:
                names = params.goniometer.names
                if len(names) != len(axes):
                    raise RuntimeError("Number of names must match axes")
            else:
                names = flex.std_string([""] * len(axes))

            # Create the scan axis
            if params.goniometer.scan_axis is not None:
                scan_axis = params.goniometer.scan_axis
            else:
                scan_axis = 0

            # Create the model
            goniometer = MultiAxisGoniometer(axes, angles, names, scan_axis)
        else:
            goniometer = reference

            # Set the axes
            if params.goniometer.axes is not None:
                axes = flex.vec3_double(
                    params.goniometer.axes[i * 3 : (i * 3) + 3]
                    for i in range(len(params.goniometer.axes) // 3)
                )
                if len(goniometer.get_axes()) != len(axes):
                    raise RuntimeError(
                        "Number of axes must match the current goniometer (%s)"
                        % len(goniometer.get_axes())
                    )
                goniometer.set_axes(axes)

            # Invert rotation axis
            if params.goniometer.invert_rotation_axis is True:
                axes = flex.vec3_double(
                    [[-x for x in v] for v in goniometer.get_axes()]
                )
                goniometer.set_axes(axes)

            # Set the angles
            if params.goniometer.angles is not None:
                if len(goniometer.get_angles()) != len(params.goniometer.angles):
                    raise RuntimeError(
                        "Number of angles must match the current goniometer (%s)"
                        % len(goniometer.get_angles())
                    )
                goniometer.set_angles(params.goniometer.angles)

            # Set the namess
            if params.goniometer.names is not None:
                if len(goniometer.get_names()) != len(params.goniometer.names):
                    raise RuntimeError(
                        "Number of names must match the current goniometer (%s)"
                        % len(goniometer.get_names())
                    )
                goniometer.set_names(params.goniometer.names)

            # Set the scan axis
            if params.goniometer.scan_axis is not None:
                raise RuntimeError("Can not override scan axis")

        return goniometer

    @staticmethod
    def from_phil(params, reference=None):
        """
        Convert the phil parameters into a goniometer model

        """
        if params.goniometer.axis and params.goniometer.axes:
            raise ValueError("Only one of axis or axes should be set")
        if reference is not None:
            if isinstance(reference, MultiAxisGoniometer):
                goniometer = GoniometerFactory.multi_axis_goniometer_from_phil(
                    params, reference
                )
            else:
                goniometer = GoniometerFactory.single_axis_goniometer_from_phil(
                    params, reference
                )
        else:
            if params.goniometer.axes and len(params.goniometer.axes) > 3:
                goniometer = GoniometerFactory.multi_axis_goniometer_from_phil(params)
            elif params.goniometer.axis or params.goniometer.axes:
                goniometer = GoniometerFactory.single_axis_goniometer_from_phil(params)
            else:
                return None
        return goniometer

    @staticmethod
    def from_dict(d, t=None):
        """Convert the dictionary to a goniometer model

        Params:
            d The dictionary of parameters
            t The template dictionary to use

        Returns:
            The goniometer model
        """
        if d is None and t is None:
            return None
        joint = t.copy() if t else {}
        joint.update(d)

        # Create the model from the joint dictionary
        if {"axes", "angles", "scan_axis"}.issubset(joint):
            return MultiAxisGoniometer.from_dict(joint)
        return Goniometer.from_dict(joint)

    @staticmethod
    def make_goniometer(rotation_axis, fixed_rotation):
        return Goniometer(
            tuple(map(float, rotation_axis)), tuple(map(float, fixed_rotation))
        )

    @staticmethod
    def make_kappa_goniometer(alpha, omega, kappa, phi, direction, scan_axis):
        omega_axis = (1, 0, 0)
        phi_axis = (1, 0, 0)

        c = math.cos(alpha * math.pi / 180)
        s = math.sin(alpha * math.pi / 180)
        if direction == "+y":
            kappa_axis = (c, s, 0.0)
        elif direction == "+z":
            kappa_axis = (c, 0.0, s)
        elif direction == "-y":
            kappa_axis = (c, -s, 0.0)
        elif direction == "-z":
            kappa_axis = (c, 0.0, -s)
        else:
            raise RuntimeError("Invalid direction")

        if scan_axis == "phi":
            scan_axis = 0
        else:
            scan_axis = 2

        axes = flex.vec3_double((phi_axis, kappa_axis, omega_axis))
        angles = flex.double((phi, kappa, omega))
        names = flex.std_string(("PHI", "KAPPA", "OMEGA"))
        return GoniometerFactory.make_multi_axis_goniometer(
            axes, angles, names, scan_axis
        )

    @staticmethod
    def make_multi_axis_goniometer(axes, angles, names, scan_axis):
        return MultiAxisGoniometer(axes, angles, names, scan_axis)

    @staticmethod
    def single_axis():
        """Construct a single axis goniometer which is canonical in the
        CBF reference frame."""

        axis = (1, 0, 0)
        fixed = (1, 0, 0, 0, 1, 0, 0, 0, 1)

        return GoniometerFactory.make_goniometer(axis, fixed)

    @staticmethod
    def single_axis_reverse():
        """Construct a single axis goniometer which is canonical in the
        CBF reference frame, but reversed in rotation."""

        axis = (-1, 0, 0)
        fixed = (1, 0, 0, 0, 1, 0, 0, 0, 1)

        return GoniometerFactory.make_goniometer(axis, fixed)

    @staticmethod
    def known_axis(axis):
        """Return an goniometer instance for a known rotation axis, assuming
        that nothing is known about the fixed element of the rotation axis."""

        assert len(axis) == 3

        fixed = (1, 0, 0, 0, 1, 0, 0, 0, 1)

        return Goniometer(axis, fixed)

    @staticmethod
    def kappa(alpha, omega, kappa, phi, direction, scan_axis):
        """Return a kappa goniometer where omega is the primary axis (i,e.
        aligned with X in the CBF coordinate frame) and has the kappa arm
        with angle alpha attached to it, aligned with -z, +y, +z or -y at
        omega = 0, that being the direction, which in turn has phi fixed to it
        which should initially be coincident with omega. We also need to know
        which axis is being used for the scan i.e. phi or omega. All angles
        should be given in degrees. This will work by first constructing the
        rotation axes and then composing them to the scan axis and fixed
        component of the rotation."""

        return GoniometerFactory.make_kappa_goniometer(
            alpha, omega, kappa, phi, direction, scan_axis
        )

    @staticmethod
    def multi_axis(axes, angles, names, scan_axis):
        """ """

        return GoniometerFactory.make_multi_axis_goniometer(
            axes, angles, names, scan_axis
        )

    @staticmethod
    def imgCIF(cif_file):
        """Initialize a goniometer model from an imgCIF file."""

        # FIXME in here work out how to get the proper setting matrix if != 1

        cbf_handle = pycbf.cbf_handle_struct()
        cbf_handle.read_file(cif_file.encode(), pycbf.MSG_DIGEST)

        return GoniometerFactory.imgCIF_H(cbf_handle)

    @staticmethod
    def imgCIF_H(cbf_handle):
        """Initialize a goniometer model from an imgCIF file handle, where
        it is assumed that the file has already been read."""

        # find the goniometer axes and dependencies
        dependants = {}
        axis_vectors = {}
        angles = {}
        scan_axis = None
        scan_axis_reversed = False
        cbf_handle.find_category(b"axis")
        for i in range(cbf_handle.count_rows()):
            cbf_handle.find_column(b"equipment")
            if cbf_handle.get_value() == b"goniometer":
                cbf_handle.find_column(b"id")
                axis_name = cbf_handle.get_value().decode()
                axis_vector = []
                for i in range(3):
                    cbf_handle.find_column(b"vector[%i]" % (i + 1))
                    axis_vector.append(float(cbf_handle.get_value()))
                axis_vectors[axis_name] = axis_vector
                cbf_handle.find_column(b"depends_on")
                dependants[cbf_handle.get_value().decode()] = axis_name
            cbf_handle.next_row()

        # find the starting angles of each goniometer axis and figure out which one
        # is the scan axis (i.e. non-zero angle_increment)
        cbf_handle.find_category(b"diffrn_scan_axis")
        for i in range(cbf_handle.count_rows()):
            cbf_handle.find_column(b"axis_id")
            axis_name = cbf_handle.get_value().decode()
            if axis_name not in axis_vectors:
                cbf_handle.next_row()
                continue
            cbf_handle.find_column(b"angle_start")
            axis_angle = float(cbf_handle.get_value())
            cbf_handle.find_column(b"angle_increment")
            increment = float(cbf_handle.get_value())
            angles[axis_name] = axis_angle
            if increment != 0:
                if scan_axis:
                    raise ValueError(
                        "More than one scan axis is defined: not currently supported."
                    )
                scan_axis = axis_name
                if increment < 0:
                    scan_axis_reversed = True
            cbf_handle.next_row()
        if not len(axis_vectors) == len(angles):
            raise ValueError(
                "The number of goniometer axes specified in the 'axis' category of the "
                "raw data does not match the number specified in the "
                "'diffrn_scan_axis' category."
            )

        # figure out the order of the axes from the depends_on values
        ordered_axes = []
        axis = dependants.get(".")
        while axis:
            ordered_axes.append(axis)
            axis = dependants.get(axis)

        # multi-axis gonio requires axes in order as viewed from crystal to gonio base
        # i.e. the reverse of the order we have from cbf header
        ordered_axes = ordered_axes[::-1]
        axis_vectors = flex.vec3_double(axis_vectors[axis] for axis in ordered_axes)
        angles = flex.double(angles[axis] for axis in ordered_axes)
        # If no scan_axis, probably a still shot ⇒ scan axis arbitrary, set to 0.
        scan_axis = ordered_axes.index(scan_axis) if scan_axis else 0

        # invert axis since the axis _values_ will be inverted in scan.py::imgCIF_H
        if scan_axis_reversed:
            a0, a1, a2 = axis_vectors[scan_axis]
            axis_vectors[scan_axis] = (-a0, -a1, -a2)

        # construct a multi-axis goniometer
        gonio = GoniometerFactory.multi_axis(
            axis_vectors, angles, flex.std_string(ordered_axes), scan_axis
        )
        return gonio



from __future__ import annotations

import os

from dxtbx.serialize.filename import resolve_path


def test_resolve_path(monkeypatch):
    monkeypatch.setenv("HELLO_WORLD", "EXPANDED")
    new_path = os.path.join("~", "$HELLO_WORLD", "path")
    path = resolve_path(new_path)
    assert path == os.path.join(os.path.expanduser("~"), "EXPANDED", "path")
    new_path = os.path.join("$HELLO_WORLD", "path")
    path = resolve_path(new_path)
    assert path == os.path.abspath(os.path.join("EXPANDED", "path"))



#
# See https://github.com/dials/dials/wiki/pytest for documentation on how to
# write and run pytest tests, and an overview of the available features.
#

from __future__ import annotations

import os
import socket

import pytest

collect_ignore = []


def dials_regression_path():
    """Return the absolute path to the dials_regression module as a string.
    This function is used directly by tests/test_regression_images.py"""

    if "DIALS_REGRESSION" in os.environ:
        return os.environ["DIALS_REGRESSION"]

    try:
        import dials_regression as dr

        return os.path.abspath(os.path.dirname(dr.__file__))
    except ImportError:
        pass  # dials_regression not configured

    # Check if we are in a known location
    reference_copy = (
        "/dls/science/groups/scisoft/DIALS/repositories/git-reference/dials_regression"
    )
    if (
        os.name == "posix"
        and socket.gethostname().endswith(".diamond.ac.uk")
        and os.path.exists(reference_copy)
    ):
        return reference_copy


@pytest.fixture(scope="session")
def dials_regression():
    """Return the absolute path to the dials_regression module as a string.
    Skip the test if dials_regression is not available."""
    d_r = dials_regression_path()
    if d_r:
        return d_r
    pytest.skip("dials_regression required for this test")


def pytest_addoption(parser):
    """Add '--regression' options to pytest."""
    try:
        parser.addoption(
            "--regression",
            action="store_true",
            default=False,
            help="run (time-intensive) regression tests",
        )
    except ValueError:
        # Thrown in case the command line option is already defined
        pass


def pytest_collection_modifyitems(config, items):
    """Tests marked as regression are only run with --regression."""
    if not config.getoption("--regression"):
        skip_regression = pytest.mark.skip(reason="Test only runs with --regression")
        for item in items:
            if "regression" in item.keywords:
                item.add_marker(skip_regression)



"""
Base implementation of miniCBF format - as used with Dectris detectors -
this will read the header and populate a dictionary of the keyword / value
pairs.
"""


from __future__ import annotations

import binascii
import datetime
import os
import pathlib
import sys

import pycbf

from boost_adaptbx.boost.python import streambuf
from cctbx import factor_ev_angstrom
from cctbx.eltbx import attenuation_coefficient
from iotbx.detectors.pilatus_minicbf import PilatusImage
from scitbx.array_family import flex

from dxtbx.ext import read_int32, uncompress
from dxtbx.format.FormatCBF import FormatCBF
from dxtbx.format.FormatCBFMiniPilatusHelpers import get_pilatus_timestamp
from dxtbx.format.FormatCBFMultiTile import cbf_wrapper
from dxtbx.model import ParallaxCorrectedPxMmStrategy, SimplePxMmStrategy

dxtbx_overload_scale = float(os.getenv("DXTBX_OVERLOAD_SCALE", "1"))


class FormatCBFMini(FormatCBF):
    """An image reading class for mini CBF format images i.e. those from
    Dectris, which will read the header into a dictionary."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an CBF format image, i.e. we can
        make sense of it."""

        header = FormatCBF.get_cbf_header(image_file)

        if "_diffrn.id" in header and "_diffrn_source" in header:
            return False

        def one_of_these_in(record):
            these = [
                "PILATUS",
                "SLS",
                "SSRL",
                "?",
                "XDS special",
                "GENERIC_MINI",  # intended for simulated PAD data, non-Pilatus array size
            ]
            for convention in these:
                if convention in record:
                    return True
            return False

        for record in header.split("\n"):
            if "_array_data.header_convention" in record and one_of_these_in(record):
                return True
            if "# Detector" in record and "PILATUS" in record:  # CBFlib v0.8.0 allowed
                return True
            if "# Detector" in record and "ADSC" in record and "HF-4M" in header:
                return True

        return False

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file."""

        self._raw_data = None
        super().__init__(image_file, **kwargs)

    @staticmethod
    def _get_timestamp_from_raw_header(
        header: str | list[str],
    ) -> datetime.datetime | None:
        """Given a raw header, or lines from, attempt to extract the timestamp field"""
        if isinstance(header, str):
            header = header.splitlines()
        timestamp = None
        for record in header:
            if len(record[1:].split()) <= 2 and record.count(":") == 2:
                timestamp = datetime.datetime.fromisoformat(record[1:].strip())
                break
        return timestamp

    def _start(self):
        """Open the image file, read the image header, copy it into a
        dictionary for future reference."""

        super()._start()
        cif_header = FormatCBF.get_cbf_header(self._image_file)

        self._cif_header_dictionary = {}

        for record in cif_header.split("\n"):
            if record[:1] != "#":
                continue

            if len(record[1:].split()) <= 2 and record.count(":") == 2:
                self._cif_header_dictionary["timestamp"] = record[1:].strip()
                continue

            tokens = record.replace("=", "").replace(":", "").split()
            self._cif_header_dictionary[tokens[1]] = " ".join(tokens[2:])

        for record in self._mime_header.split("\n"):
            if not record.strip():
                continue
            token, value = record.split(":")
            self._cif_header_dictionary[token.strip()] = value.strip()

    def _detector(self):
        """Return a model for a simple detector, presuming no one has
        one of these on a two-theta stage. Assert that the beam centre is
        provided in the Mosflm coordinate frame."""

        distance = float(self._cif_header_dictionary["Detector_distance"].split()[0])

        beam_xy = (
            self._cif_header_dictionary["Beam_xy"]
            .replace("(", "")
            .replace(")", "")
            .replace(",", "")
            .split()[:2]
        )

        wavelength = float(self._cif_header_dictionary["Wavelength"].split()[0])

        beam_x, beam_y = map(float, beam_xy)

        pixel_xy = (
            self._cif_header_dictionary["Pixel_size"]
            .replace("m", "")
            .replace("x", "")
            .split()
        )

        pixel_x, pixel_y = map(float, pixel_xy)

        if "Silicon" in self._cif_header_dictionary:
            thickness = (
                float(self._cif_header_dictionary["Silicon"].split()[2]) * 1000.0
            )
            material = "Si"
            sensor = "PAD"
        elif "CdTe" in self._cif_header_dictionary:
            thickness = float(self._cif_header_dictionary["CdTe"].split()[2]) * 1000.0
            material = "CdTe"
            sensor = "PAD"
        elif "CCD" in self._cif_header_dictionary:
            thickness = 0
            material = None
            sensor = "CCD"
        else:
            thickness = 0
            material = None
            sensor = None

        nx = int(self._cif_header_dictionary["X-Binary-Size-Fastest-Dimension"])
        ny = int(self._cif_header_dictionary["X-Binary-Size-Second-Dimension"])

        overload = dxtbx_overload_scale * int(
            self._cif_header_dictionary["Count_cutoff"].split()[0]
        )
        minimum_trusted_value = 0

        if material is not None:
            # take into consideration here the thickness of the sensor also the
            # wavelength of the radiation (which we have in the same file...)
            table = attenuation_coefficient.get_table(material)
            mu = table.mu_at_angstrom(wavelength) / 10.0
            t0 = thickness
            px_mm = ParallaxCorrectedPxMmStrategy(mu, t0)
        else:
            px_mm = SimplePxMmStrategy()

        detector = self._detector_factory.simple(
            sensor,
            distance * 1000.0,
            (beam_x * pixel_x * 1000.0, beam_y * pixel_y * 1000.0),
            "+x",
            "-y",
            (1000 * pixel_x, 1000 * pixel_y),
            (nx, ny),
            (minimum_trusted_value, overload),
            [],
            px_mm=px_mm,
        )

        if material is not None:
            detector[0].set_thickness(thickness)
            detector[0].set_material(material)
            detector[0].set_mu(mu)

        return detector

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer. This should
        probably be checked against the image header, though for miniCBF
        there are limited options for this."""

        #  if "Phi" in self._cif_header_dictionary:
        #      phi_value = float(self._cif_header_dictionary["Phi"].split()[0])

        return self._goniometer_factory.single_axis()

    def _beam(self):
        """Return a simple model for the beam."""

        wavelength = float(self._cif_header_dictionary["Wavelength"].split()[0])

        beam = self._beam_factory.simple(wavelength)

        try:
            flux = float(self._cif_header_dictionary["Flux"].split()[0])
            beam.set_flux(flux)
        except (IndexError, KeyError):
            pass

        try:
            transmission = float(self._cif_header_dictionary["Transmission"].split()[0])
            beam.set_transmission(transmission)
        except (IndexError, KeyError):
            pass

        return beam

    def _scan(self):
        exposure_time = float(self._cif_header_dictionary["Exposure_period"].split()[0])
        osc_start = float(self._cif_header_dictionary["Start_angle"].split()[0])
        osc_range = float(self._cif_header_dictionary["Angle_increment"].split()[0])

        if "timestamp" in self._cif_header_dictionary:
            timestamp = get_pilatus_timestamp(self._cif_header_dictionary["timestamp"])
        else:
            timestamp = 0.0

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, timestamp
        )

    def _read_cbf_image(self):
        start_tag = binascii.unhexlify("0c1a04d5")

        with self.open_file(self._image_file, "rb") as fh:
            data = fh.read()
        data_offset = data.find(start_tag) + 4
        cbf_header = self._parse_cbf_header(
            data[: data_offset - 4].decode("ascii", "ignore")
        )

        if cbf_header["byte_offset"]:
            pixel_values = uncompress(
                packed=data[data_offset : data_offset + cbf_header["size"]],
                fast=cbf_header["fast"],
                slow=cbf_header["slow"],
            )
        elif cbf_header["no_compression"]:
            assert len(self.get_detector()) == 1
            with self.open_file(self._image_file) as f:
                f.read(data_offset)
                pixel_values = read_int32(streambuf(f), cbf_header["length"])
            pixel_values.reshape(flex.grid(cbf_header["slow"], cbf_header["fast"]))

        else:
            raise ValueError(
                "Compression of type other than byte_offset or none is not supported (contact authors)"
            )

        return pixel_values

    def get_raw_data(self):
        if self._raw_data is None:
            data = self._read_cbf_image()
            self._raw_data = data

        return self._raw_data

    def detectorbase_start(self):
        self.detectorbase = PilatusImage(self._image_file)
        self.detectorbase.readHeader()  # necessary for LABELIT

    @staticmethod
    def as_file(
        detector,
        beam,
        gonio,
        scan,
        data,
        path,
        header_convention="GENERIC_MINI",
        det_type="GENERIC",
    ):
        """Note to developers: first attempt to write a miniCBF given a dxtbx-style experiment,
        But fields are not filled rigorously as in cbflib/src/cbf_minicbf_header.c
        Present code does not account for:
          Pilatus model number and serial number
          Data collection date and time
          Sensor material
          Dead time (exposure time - exposure period)
          Highest trusted value
          Energy threshold for photon counting
          Gain
          Bad pixels
          Auxiliary files (bad pixel mask, flat field, trim, image path)
          Detector not normal to beam
        """
        path = pathlib.Path(path)
        cbf = cbf_wrapper()
        cbf_root = path.with_suffix(".cbf")
        cbf.new_datablock(path.stem.encode())

        """Data items in the ARRAY_DATA category are the containers for the array data
        items described in the category ARRAY_STRUCTURE."""
        cbf.add_category("array_data", ["header_convention", "header_contents", "data"])

        # get pixel info out of detector object
        panel = detector[0]
        pixel_xy = panel.get_pixel_size()
        pixel_x_microns = pixel_xy[0] * 1000
        pixel_y_microns = pixel_xy[1] * 1000

        # make sure we get the right units
        thickness_meters = panel.get_thickness() / 1000.0

        material = panel.get_material()
        if material == "Si":
            sensor = "Silicon"
        elif material == "CdTe":
            sensor = "CdTe"
        elif panel.get_type() == "SENSOR_CCD":
            sensor = "CCD"
        else:
            sensor = "unknown"

        # maybe someday more people will do this
        flux = beam.get_flux()
        transmission = beam.get_transmission()
        polarization_fraction = beam.get_polarization_fraction()

        # get exposure information
        if scan is None:
            exposure_period = phi_start = osc = 0
        else:
            exposure_period = scan.get_exposure_times()[0]
            phi_start = scan.get_oscillation()[0]
            osc = scan.get_oscillation()[1]

        # automatically account for read-out time?
        # exposure_time = exposure_period-0.00203  # this is appropriate for Pilatus3 S
        exposure_time = exposure_period  # simulation is a perfect detector

        tau = 0  # assume simulation is a perfect detector with no pile-up error
        trusted_range = detector[0].get_trusted_range()
        count_cutoff = trusted_range[1]

        wavelength = beam.get_wavelength()  # get the wavelength in the conventional way
        energy = factor_ev_angstrom / wavelength
        threshold = energy / 2  # presume normal data collection convention

        bad_pixels = 0  # maybe get this from negative pixel values?

        assert len(detector) == 1, "only single-panel detectors supported"
        distance_meters = detector[0].get_distance() / 1000

        # fixed now, no longer backwards
        beam_center = detector[0].get_beam_centre_px(beam.get_s0())
        ORGX, ORGY = beam_center

        cbf.add_row(
            [
                header_convention,
                """
# Detector: %(det_type)s, S/N 60-0000
# 1972-01-01T00:00:00.000
# Pixel_size %(pixel_x_microns)ge-6 m x %(pixel_x_microns)ge-6 m
# %(sensor)s sensor, thickness %(thickness_meters)f m
# Exposure_time %(exposure_time).7f s
# Exposure_period %(exposure_period).7f s
# Tau = %(tau)f s
# Count_cutoff %(count_cutoff)d counts
# Threshold_setting: %(threshold)d eV
# Gain_setting: autog (vrf = 1.000)
# N_excluded_pixels = %(bad_pixels)d
# Excluded_pixels: badpix_mask.tif
# Flat_field: (nil)
# Trim_file: (nil)
# Image_path: /ramdisk/
# Wavelength %(wavelength)g A
# Detector_distance %(distance_meters)g m
# Beam_xy (%(ORGX)g, %(ORGY)g) pixels
# Flux %(flux)g
# Filter_transmission %(transmission).4f
# Start_angle %(phi_start).4f deg.
# Angle_increment %(osc).4f deg.
# Polarization %(polarization_fraction).3f
# Detector_2theta 0.0000 deg.
# Kappa 0.0000 deg.
# Phi %(phi_start).4f deg.
"""
                % locals(),
            ]
        )

        binary_id = 1
        focus = data.focus()
        data2 = data.copy_to_byte_str()
        elements = len(data)
        byteorder = b"little_endian"
        dimfast = focus[1]
        dimmid = focus[0]
        dimslow = 1
        padding = 0
        elsize = 4
        elsigned = 1

        cbf.set_integerarray_wdims_fs(
            pycbf.CBF_BYTE_OFFSET,
            binary_id,
            data2,
            elsize,
            elsigned,
            elements,
            byteorder,
            dimfast,
            dimmid,
            dimslow,
            padding,
        )

        cbf.write_widefile(
            str(cbf_root).encode(),
            pycbf.CBF,
            pycbf.MIME_HEADERS | pycbf.MSG_DIGEST | pycbf.PAD_4K,
            0,
        )


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMini.understand(arg))



from __future__ import annotations

import h5py
import nxmx

from scitbx.array_family import flex

from dxtbx.format import nexus
from dxtbx.format.FormatNXmx import FormatNXmx
from dxtbx.masking import mask_untrusted_circle, mask_untrusted_polygon
from dxtbx.model import SimplePxMmStrategy
from dxtbx.model.beam import Probe


class FormatNXmxED(FormatNXmx):
    """Format class for 3DED/MicroED images that have been converted to NXmx
    by nexgen"""

    @staticmethod
    def understand(image_file):
        with h5py.File(image_file) as handle:
            if FormatNXmxED.get_source_probe(handle).lower() != "electron":
                return False
        return True

    @staticmethod
    def get_source_probe(handle) -> str:
        try:
            probe = handle["/entry/source/probe"][()]
            probe = nexus.h5str(probe)
        except KeyError:
            probe = ""
        return probe

    def _detector(self):
        """Overridden to change px-to-mm and QE corrections from their X-ray
        defaults"""

        # DECTRIS recommend a paper by Fernandez-Perez et al 2021
        # (https://doi.org/10.1088/1748-0221/16/10/P10034) that provides some
        # characterisation of the detector response to fast electrons. Fig 1
        # shows there is essentially no transmission through the sensor layer.
        # Most charge is accumulated near the surface of the sensor, but due
        # to the random-walk effect of multiple interactions as the electron
        # slows down, there is significant spreading of the signal beyond the
        # incident pixel. The parallax-corrected px-to-mm function used for
        # X-rays is not correct for this situation, so we will set the simple
        # strategy instead.
        #
        # A second issue is that of the QE correction. As there is no
        # transmission of electrons through the sensor, the simple model using
        # a linear absorption coefficient, µ, is also not appropriate here.
        # There is loss of signal due to other effects, such as backscattering,
        # but we do not have a model for this. As a result, we shall set µ to be
        # very high to effectively nullify the correction.

        detector = super()._detector()
        for panel in detector:
            panel.set_px_mm_strategy(SimplePxMmStrategy())
            panel.set_mu(1e10)

        return detector

    def _beam(self, index=None):
        """Ensure the beam is unpolarised"""

        beam = super()._beam()
        beam.set_polarization_fraction(0.5)
        beam.set_probe(Probe.electron)

        return beam


class FormatNXmxEDeBIC(FormatNXmxED):

    """Set a mask specific for the SINGLA installed at eBIC"""

    @staticmethod
    def understand(image_file):
        with h5py.File(image_file) as handle:
            name = nxmx.h5str(FormatNXmxEDeBIC.get_instrument_name(handle))
            if name and "ebic" in name.lower():
                return True
        return False

    def _start(self):
        super()._start()

        self._extend_static_mask()

    def _extend_static_mask(self):
        """Extend the existing static mask to include the flange shadow."""

        pixel_mask = self._static_mask[0]

        # It is easier to calculate the inverted mask for the shadow of the vacuum flange
        inverted_flange_mask = flex.bool(flex.grid(pixel_mask.all()), True)
        for coord in [(201, 210), (202, 846), (836, 208), (838, 845)]:
            mask_untrusted_circle(inverted_flange_mask, coord[0], coord[1], 194)
        vertices = [
            (7, 210),
            (201, 16),
            (836, 14),
            (1028, 174),
            (1028, 886),
            (838, 1039),
            (202, 1040),
            (8, 836),
        ]
        polygon = flex.vec2_double(vertices)
        mask_untrusted_polygon(inverted_flange_mask, polygon)

        # Combine masks taking care of the inversion
        self._static_mask = ((pixel_mask & ~inverted_flange_mask),)



"""
An implementation of the CBF image reader for Pilatus images, from the Pilatus
6M SN 100 currently on Diamond I04.
"""


from __future__ import annotations

import os
import sys

from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus
from dxtbx.format.FormatStill import FormatStill


class FormatCBFMiniPilatusDLS6MSN114DMM(FormatCBFMiniPilatus, FormatStill):
    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Pilatus mini CBF format image,
        i.e. we can make sense of it."""

        if os.environ.get("I02_DMM", "0") != "1":
            return False

        header = FormatCBFMiniPilatus.get_cbf_header(image_file)

        for record in header.split("\n"):
            if (
                "# Detector" in record
                and "PILATUS" in record
                and "S/N 60-0114-F" in header
            ):
                return True

        return False

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file, including a
        proper model of the experiment."""

        FormatCBFMiniPilatus.__init__(self, image_file, **kwargs)
        FormatStill.__init__(self, image_file, **kwargs)

        return

    def _goniometer(self):
        return None

    def _scan(self):
        return None


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniPilatusDLS6MSN114DMM.understand(arg))



from __future__ import annotations

from typing import Any

import pycbf

class ImageBool:
    def __init__(self, *args, **kwargs) -> None: ...
    def append(self, *args, **kwargs) -> Any: ...
    def empty(dxtbx) -> Any: ...
    def n_tiles(dxtbx) -> Any: ...
    def tile(dxtbx, unsignedlong) -> Any: ...
    def tile_names(dxtbx) -> Any: ...
    def __getitem__(dxtbx, unsignedlong) -> Any: ...
    def __getstate__(dxtbx) -> Any: ...
    def __iter__(boost) -> Any: ...
    def __len__(dxtbx) -> Any: ...
    def __reduce__(self) -> Any: ...
    def __setstate__(dxtbx, boost) -> Any: ...

class ImageBuffer:
    def __init__(self, *args, **kwargs) -> None: ...
    def as_double(dxtbx) -> Any: ...
    def as_float(dxtbx) -> Any: ...
    def as_int(dxtbx) -> Any: ...
    def is_double(dxtbx) -> Any: ...
    def is_empty(dxtbx) -> Any: ...
    def is_float(dxtbx) -> Any: ...
    def is_int(dxtbx) -> Any: ...
    def __reduce__(self) -> Any: ...

class ImageDouble:
    def __init__(self, *args, **kwargs) -> None: ...
    def append(self, *args, **kwargs) -> Any: ...
    def empty(dxtbx) -> Any: ...
    def n_tiles(dxtbx) -> Any: ...
    def tile(dxtbx, unsignedlong) -> Any: ...
    def tile_names(dxtbx) -> Any: ...
    def __getitem__(dxtbx, unsignedlong) -> Any: ...
    def __getstate__(dxtbx) -> Any: ...
    def __iter__(boost) -> Any: ...
    def __len__(dxtbx) -> Any: ...
    def __reduce__(self) -> Any: ...
    def __setstate__(dxtbx, boost) -> Any: ...

class ImageInt:
    def __init__(self, *args, **kwargs) -> None: ...
    def append(self, *args, **kwargs) -> Any: ...
    def empty(dxtbx) -> Any: ...
    def n_tiles(dxtbx) -> Any: ...
    def tile(dxtbx, unsignedlong) -> Any: ...
    def tile_names(dxtbx) -> Any: ...
    def __getitem__(dxtbx, unsignedlong) -> Any: ...
    def __getstate__(dxtbx) -> Any: ...
    def __iter__(boost) -> Any: ...
    def __len__(dxtbx) -> Any: ...
    def __reduce__(self) -> Any: ...
    def __setstate__(dxtbx, boost) -> Any: ...

class ImageTileBool:
    def __init__(self, *args, **kwargs) -> None: ...
    def data(dxtbx) -> Any: ...
    def empty(dxtbx) -> Any: ...
    def name(dxtbx) -> Any: ...
    def __getinitargs__(dxtbx) -> Any: ...
    def __reduce__(self) -> Any: ...

class ImageTileDouble:
    def __init__(self, *args, **kwargs) -> None: ...
    def data(dxtbx) -> Any: ...
    def empty(dxtbx) -> Any: ...
    def name(dxtbx) -> Any: ...
    def __getinitargs__(dxtbx) -> Any: ...
    def __reduce__(self) -> Any: ...

class ImageTileInt:
    def __init__(self, *args, **kwargs) -> None: ...
    def data(dxtbx) -> Any: ...
    def empty(dxtbx) -> Any: ...
    def name(dxtbx) -> Any: ...
    def __getinitargs__(dxtbx) -> Any: ...
    def __reduce__(self) -> Any: ...

def cbf_read_buffer(
    handle: pycbf.cbf_handle_struct, buffer: bytes, flags: int
) -> None: ...



from __future__ import annotations

import argparse

from scitbx.array_family import flex

import dxtbx.format.Registry
import dxtbx.util
from dxtbx.format.FormatMultiImage import FormatMultiImage


def print_header(filenames):
    # this will do the lookup for every frame - this is strictly not needed
    # if all frames are from the same instrument
    for arg in filenames:
        print(f"=== {arg} ===")
        format_class = dxtbx.format.Registry.get_format_class_for_file(arg)
        if not format_class:
            print(f"No format class found that can understand {arg}")
            continue
        print(f"Using header reader: {format_class.__name__}")
        i = format_class(arg)
        beam = i.get_beam()
        goniometer = i.get_goniometer()
        detector = i.get_detector()
        scan = i.get_scan()
        if beam is None:
            print("No beam model found")
        else:
            print(beam)
        if detector is None:
            print("No detector model found")
        else:
            print(detector)
        if goniometer is None:
            print("No goniometer model found")
        else:
            print(goniometer)
        if scan is None:
            print("No scan model found")
        else:
            print(scan)

        if not issubclass(format_class, FormatMultiImage):
            try:
                raw_data = i.get_raw_data()
                if not isinstance(raw_data, tuple):
                    raw_data = (raw_data,)
                d = [p.as_1d() for p in raw_data]
                print("Total Counts: %d" % sum([flex.sum(p.select(p >= 0)) for p in d]))
            except AttributeError:
                print("Could not read image data")


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = argparse.ArgumentParser(description="Print headers for images")
    parser.add_argument("image_files", nargs="+", metavar="FILE", help="Image files")
    options = parser.parse_args(args)
    print_header(options.image_files)


if __name__ == "__main__":
    run()



"""An implementation of the SMV image reader for CMOS1 images, from ALS 4.2.2"""


from __future__ import annotations

import calendar
import sys
import time

from boost_adaptbx.boost.python import streambuf
from scitbx import matrix
from scitbx.array_family import flex

from dxtbx.ext import read_uint16
from dxtbx.format.FormatSMV import FormatSMV


class FormatSMVCMOS1(FormatSMV):
    """A class for reading SMV format CMOS1 images. 'Abstract' class."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like a CMOS1 d*TREK SMV format image,
        i.e. we can make sense of it. Essentially that will be if it contains
        all of the keys we are looking for."""

        size, header = FormatSMV.get_smv_header(image_file)

        wanted_header_items = [
            "DETECTOR_NUMBER",
            "DETECTOR_NAMES",
            "BYTE_ORDER",
            "DIM",
            "SIZE1",
            "SIZE2",
            "Data_type",
        ]

        if any(item not in header for item in wanted_header_items):
            return False

        detector_prefixes = header["DETECTOR_NAMES"].split()

        if len(detector_prefixes) != 1:
            return False

        detector_prefix = detector_prefixes[0]

        more_wanted_header_items = [
            "DETECTOR_DIMENSIONS",
            "DETECTOR_SIZE",
            "DETECTOR_VECTORS",
            "GONIO_NAMES",
            "GONIO_UNITS",
            "GONIO_VALUES",
            "GONIO_VECTORS",
        ]

        if any(
            f"{detector_prefix}{item}" not in header
            for item in more_wanted_header_items
        ):
            return False

        det_desc = "%sDETECTOR_DESCRIPTION" % detector_prefix
        if "CMOS-1" in header.get(det_desc, ""):
            return True

        return False

    def __init__(self, image_file, **kwargs):
        """Initialise the image structure from the given file, including a
        proper model of the experiment."""

        super().__init__(image_file, **kwargs)
        detector_prefixes = self._header_dictionary["DETECTOR_NAMES"].split()
        self._prefix = detector_prefixes[0]

    def _start(self):
        super()._start()
        self._header_size = int(self._header_dictionary["HEADER_BYTES"])

    def _goniometer(self):
        axis = tuple(map(float, self._header_dictionary["ROTATION_VECTOR"].split()))

        return self._goniometer_factory.known_axis(axis)

    def _detector(self):
        detector_name = self._header_dictionary["DETECTOR_NAMES"].split()[0].strip()

        detector_axes = self.get_detector_axes(detector_name)
        detector_fast = matrix.col(tuple(detector_axes[:3]))
        detector_slow = matrix.col(tuple(detector_axes[3:]))

        beam_pixels = self.get_beam_pixels(detector_name)
        pixel_size = self.get_pixel_size(detector_name)
        image_size = self.get_image_size(detector_name)

        detector_origin = -(
            beam_pixels[0] * pixel_size[0] * detector_fast
            + beam_pixels[1] * pixel_size[1] * detector_slow
        )

        gonio_axes = self.get_gonio_axes(detector_name)
        gonio_values = self.get_gonio_values(detector_name)
        gonio_units = self._header_dictionary["%sGONIO_UNITS" % detector_name].split()
        gonio_num_axes = int(
            self._header_dictionary["%sGONIO_NUM_VALUES" % detector_name]
        )

        rotations = []
        translations = []

        for j, unit in enumerate(gonio_units):
            axis = matrix.col(gonio_axes[3 * j : 3 * (j + 1)])
            if unit == "deg":
                rotations.append(
                    axis.axis_and_angle_as_r3_rotation_matrix(gonio_values[j], deg=True)
                )
                translations.append(matrix.col((0.0, 0.0, 0.0)))
            elif unit == "mm":
                rotations.append(
                    matrix.sqr((1.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 1.0))
                )
                translations.append(gonio_values[j] * axis)
            else:
                raise RuntimeError("unknown axis unit %s" % unit)

        rotations.reverse()
        translations.reverse()

        for j in range(gonio_num_axes):
            detector_fast = rotations[j] * detector_fast
            detector_slow = rotations[j] * detector_slow
            detector_origin = rotations[j] * detector_origin
            detector_origin = translations[j] + detector_origin

        max_trusted_value = int(float(self._header_dictionary["SATURATED_VALUE"]))
        min_trusted_value = 0

        return self._detector_factory.complex(
            "CCD",
            detector_origin.elems,
            detector_fast.elems,
            detector_slow.elems,
            pixel_size,
            image_size,
            (min_trusted_value, max_trusted_value),
        )

    def _beam(self):
        beam_direction = self.get_beam_direction()
        p_fraction, p_plane = self.get_beam_polarization()

        wavelength = float(self._header_dictionary["WAVELENGTH"])

        return self._beam_factory.complex(
            beam_direction, p_fraction, p_plane, wavelength
        )

    def _scan(self):
        rotation = self.get_rotation()

        time_record = self._header_dictionary["DATE"]
        date_record, ms = time_record.split(".")
        epoch = calendar.timegm(time.strptime(date_record, "%a %b %d %Y %H:%M:%S"))
        epoch += 0.001 * int(ms)

        exposure_time = rotation[3]
        osc_start = rotation[0]
        osc_range = rotation[2]

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, epoch
        )

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array of integers.)"""

        # currently have no non-little-endian machines...
        assert len(self.get_detector()) == 1
        image_size = self.get_detector()[0].get_image_size()
        with self.open_file(self._image_file) as fh:
            fh.seek(self._header_size)
            raw_data = read_uint16(streambuf(fh), int(image_size[0] * image_size[1]))
        raw_data.reshape(flex.grid(image_size[1], image_size[0]))

        return raw_data


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatSMVCMOS1.understand(arg))



from __future__ import annotations

from dxtbx.format.FormatCBFMiniPilatus import FormatCBFMiniPilatus


class FormatCBFMiniPilatusSPring8_6MSN125(FormatCBFMiniPilatus):
    """A class for reading mini CBF format Pilatus images for 6M SN 125, normally
    at Spring8 BL41XU"""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Pilatus mini CBF format image,
        i.e. we can make sense of it."""

        header = FormatCBFMiniPilatus.get_cbf_header(image_file)

        for record in header.split("\n"):
            if (
                "# Detector" in record
                and "PILATUS" in record
                and "S/N 60-0125" in header
            ):
                return True

        return False

    def _goniometer(self):
        """Return a model for a simple single-axis reversed direction goniometer."""

        return self._goniometer_factory.single_axis_reverse()



"""
ADSC SMV Format for Q315 SN 915, installed at BL38B1 at SPring-8. Resembles
FormatSMVADSCSN920 but returns a reverse phi goniometer
"""

from __future__ import annotations

from dxtbx.format.FormatSMVADSCSN import FormatSMVADSCSN


class FormatSMVADSCSN915(FormatSMVADSCSN):
    """A class for reading SMV format ADSC images, and correctly constructing
    a model for the experiment from this, for instrument number 915."""

    @staticmethod
    def understand(image_file):
        """Check to see if this is ADSC SN 915."""

        # check this is detector serial number 915

        size, header = FormatSMVADSCSN.get_smv_header(image_file)

        return int(header["DETECTOR_SN"]) == 915

    def _goniometer(self):
        """Return a model for a simple single-axis goniometer with reversed
        direction."""

        return self._goniometer_factory.single_axis_reverse()



from __future__ import annotations

import os

import pycbf

import libtbx.phil
from scitbx.array_family import flex

from dxtbx.model.scan_helpers import scan_helper_image_files

try:
    from ..dxtbx_model_ext import Scan
except ModuleNotFoundError:
    from dxtbx_model_ext import Scan  # type: ignore

scan_phil_scope = libtbx.phil.parse(
    """
  scan
    .expert_level = 1
    .short_caption = "Scan overrides"
  {

    image_range = None
      .type = ints(size=2)
      .help = "Override the image range"
      .short_caption = "Image range"

    extrapolate_scan = False
      .type = bool
      .help = "When overriding the image range, extrapolate exposure and epoch information from existing images"
      .short_caption = "Extrapolate scan"

    oscillation = None
      .type = floats(size=2)
      .help = "Override the image oscillation"
      .short_caption = "Oscillation"

    batch_offset = None
      .type = int(value_min=0)
      .help = "Override the batch offset"
      .short_caption = "Batch offset"
  }
"""
)


class ScanFactory:
    """A factory for scan instances, to help with constructing the classes
    in a set of common circumstances."""

    @staticmethod
    def from_phil(params, reference=None):
        """
        Generate a scan model from phil parameters

        """
        if reference is None:
            if params.scan.image_range is None and params.scan.oscillation is None:
                return None
            if params.scan.image_range is None:
                raise RuntimeError("No image range set")
            if params.scan.oscillation is None:
                raise RuntimeError("No oscillation set")
            scan = Scan(params.scan.image_range, params.scan.oscillation)
        else:
            scan = reference

            if params.scan.image_range is not None:
                most_recent_image_index = (
                    scan.get_image_range()[1] - scan.get_image_range()[0]
                )
                scan.set_oscillation(
                    scan.get_image_oscillation(params.scan.image_range[0])
                )
                scan.set_image_range(params.scan.image_range)
                if (
                    params.scan.extrapolate_scan
                    and (params.scan.image_range[1] - params.scan.image_range[0])
                    > most_recent_image_index
                ):
                    exposure_times = scan.get_exposure_times()
                    epochs = scan.get_epochs()
                    exposure_time = exposure_times[most_recent_image_index]
                    epoch_correction = epochs[most_recent_image_index]
                    for i in range(
                        most_recent_image_index + 1,
                        params.scan.image_range[1] - params.scan.image_range[0] + 1,
                    ):
                        exposure_times[i] = exposure_time
                        epoch_correction += exposure_time
                        epochs[i] = epoch_correction
                    scan.set_epochs(epochs)
                    scan.set_exposure_times(exposure_times)

            if params.scan.oscillation is not None:
                scan.set_oscillation(params.scan.oscillation)

        if params.scan.batch_offset is not None:
            scan.set_batch_offset(params.scan.batch_offset)

        return scan

    @staticmethod
    def from_dict(d, t=None):
        """Convert the dictionary to a scan model

        Params:
            d The dictionary of parameters
            t The template dictionary to use

        Returns:
            The scan model
        """
        if d is None and t is None:
            return None
        joint = t.copy() if t else {}
        joint.update(d)

        if not isinstance(joint["exposure_time"], list):
            joint["exposure_time"] = [joint["exposure_time"]]
        joint.setdefault("batch_offset", 0)  # backwards compatibility 20180205
        joint.setdefault("valid_image_ranges", {})  # backwards compatibility 20181113

        # Create the model from the joint dictionary
        return Scan.from_dict(joint)

    @staticmethod
    def make_scan(
        image_range, exposure_times, oscillation, epochs, batch_offset=0, deg=True
    ):
        if not isinstance(exposure_times, list):
            num_images = image_range[1] - image_range[0] + 1
            exposure_times = [exposure_times for i in range(num_images)]
        else:
            num_images = image_range[1] - image_range[0] + 1
            num_exp = len(exposure_times)
            if num_exp != num_images:
                if num_exp == 0:
                    exposure_times = [0 for i in range(num_images)]
                else:
                    exposure_times = exposure_times.extend(
                        [exposure_times[-1] for i in range(num_images - num_exp)]
                    )

        epoch_list = [epochs[j] for j in sorted(epochs)]

        return Scan(
            tuple(map(int, image_range)),
            tuple(map(float, oscillation)),
            flex.double(list(map(float, exposure_times))),
            flex.double(list(map(float, epoch_list))),
            batch_offset,
            deg,
        )

    @staticmethod
    def single_file(filename, exposure_times, osc_start, osc_width, epoch):
        """Construct an scan instance for a single image."""
        index = scan_helper_image_files.image_to_index(os.path.split(filename)[-1])
        if epoch is None:
            epoch = 0.0

        # if the oscillation width is negative at this stage it is almost
        # certainly an artefact of the omega end being 0 when the omega start
        # angle was ~ 360 so it would be ~ -360 - see dxtbx#378
        if osc_width < -180:
            osc_width += 360

        return ScanFactory.make_scan(
            (index, index), exposure_times, (osc_start, osc_width), {index: epoch}
        )

    @staticmethod
    def imgCIF(cif_file):
        """Initialize a scan model from an imgCIF file."""

        cbf_handle = pycbf.cbf_handle_struct()
        cbf_handle.read_file(cif_file, pycbf.MSG_DIGEST)

        return ScanFactory.imgCIF_H(cif_file, cbf_handle)

    @staticmethod
    def imgCIF_H(cif_file, cbf_handle):
        """Initialize a scan model from an imgCIF file handle, where it is
        assumed that the file has already been read."""

        exposure = cbf_handle.get_integration_time()
        timestamp = cbf_handle.get_timestamp()[0]

        gonio = cbf_handle.construct_goniometer()
        try:
            angles = tuple(gonio.get_rotation_range())
        except Exception as e:
            if str(e).strip() == "CBFlib Error(s): CBF_NOTFOUND":
                # probably a still shot -> no scan object
                return None
            raise

        # xia2-56 handle gracefully reverse turning goniometers - this assumes the
        # rotation axis is correctly inverted in the goniometer factory
        if angles[1] < 0:
            angles = -angles[0], -angles[1]

        index = scan_helper_image_files.image_to_index(cif_file)

        gonio.__swig_destroy__(gonio)

        return ScanFactory.make_scan(
            (index, index), exposure, angles, {index: timestamp}
        )

    @staticmethod
    def add(scans):
        """Sum a list of scans wrapping the slightly clumsy idiomatic method:
        sum(scans[1:], scans[0])."""
        return sum(scans[1:], scans[0])

    @staticmethod
    def search(filename):
        """Get a list of files which appear to match the template and
        directory implied by the input filename. This could well be used
        to get a list of image headers to read and hence construct scans
        from."""

        template, directory = scan_helper_image_files.image_to_template_directory(
            filename
        )

        indices = scan_helper_image_files.template_directory_to_indices(
            template, directory
        )

        return [
            scan_helper_image_files.template_directory_index_to_image(
                template, directory, index
            )
            for index in indices
        ]



from __future__ import annotations

import boost_adaptbx.boost.python

ext = boost_adaptbx.boost.python.import_ext("dxtbx_ext")

try:
    from .dxtbx_ext import *  # noqa: F401,F403,E402
except ModuleNotFoundError:
    from dxtbx_ext import *  # type: ignore # noqa: F401,F403,E402



"""
An implementation of the SMV image reader for ADSC images. Inherits from
FormatSMVADSC, customised for old detector on Diamond Light Source I03,
correctly accounting for the image pedestal & similar
"""

from __future__ import annotations

from dxtbx.format.FormatSMVADSCSN import FormatSMVADSCSN


class FormatSMVADSCSN920(FormatSMVADSCSN):
    """A class for reading SMV format ADSC images, and correctly constructing
    a model for the experiment from this, for instrument number 920."""

    @staticmethod
    def understand(image_file):
        """Check to see if this is ADSC SN 920."""

        # check this is detector serial number 920

        size, header = FormatSMVADSCSN.get_smv_header(image_file)

        return int(header["DETECTOR_SN"]) == 920



from __future__ import annotations

import itertools
import math
from operator import itemgetter
from typing import TYPE_CHECKING, Tuple, cast

import numpy as np

from scitbx import matrix

if TYPE_CHECKING:
    from scitbx.array_family import flex

    from dxtbx.model import Detector, Panel

try:
    import sklearn.cluster
except ImportError:
    sklearn = None

Float2 = Tuple[float, float]
Float4 = Tuple[float, float, float, float]


def read_xds_xparm(xds_xparm_file):
    """Parse the XDS XPARM file, which contains a description of the detector
    and experimental geometry, to a dictionary."""

    with open(xds_xparm_file) as fh:
        data = [float(i) for i in fh.read().split()]

    assert len(data) == 42

    starting_frame = int(data[0])
    phi_start, phi_width = data[1:3]
    axis = data[3:6]

    wavelength = data[6]
    beam = data[7:10]

    nx, ny = map(int, data[10:12])
    px, py = data[12:14]

    distance = data[14]
    ox, oy = data[15:17]

    x, y = data[17:20], data[20:23]
    normal = data[23:26]

    spacegroup = int(data[26])
    cell = data[27:33]

    a, b, c = data[33:36], data[36:39], data[39:42]

    results = {
        "starting_frame": starting_frame,
        "phi_start": phi_start,
        "phi_width": phi_width,
        "axis": axis,
        "wavelength": wavelength,
        "beam": beam,
        "nx": nx,
        "ny": ny,
        "px": px,
        "py": py,
        "distance": distance,
        "ox": ox,
        "oy": oy,
        "x": x,
        "y": y,
        "normal": normal,
        "spacegroup": spacegroup,
        "cell": cell,
        "a": a,
        "b": b,
        "c": c,
    }

    return results


def compute_frame_rotation(original, final):
    """Compute reference frame rotation to rotate from the original frame
    given by original = (x, y, z) to the to reference frame given by
    final = (_x, _y, _z). Returns M where M.x = _x etc."""

    x, y, z = original
    _x, _y, _z = final

    original_matrix = matrix.sqr(x.elems + y.elems + z.elems).transpose()
    assert (original_matrix.determinant() - 1.0) < 1.0e-7

    final_matrix = matrix.sqr(_x.elems + _y.elems + _z.elems).transpose()
    assert (final_matrix.determinant() - 1.0) < 1.0e-7

    # #1 rotate about x ^ (1, 0, 0) - if they are not coincident,
    # rotate about _x ^ _y if they are colinear but in opposite
    # directions

    if _x.angle(x) % math.pi:
        _ra_x = _x.cross(x)
        _a_x = _x.angle(x)
    elif math.fabs(_x.angle(x) - math.pi) < 1.0e-7:
        _ra_x = _x.cross(_y)
        _a_x = math.pi
    else:
        _ra_x = _x
        _a_x = 0.0

    _m_x = _ra_x.axis_and_angle_as_r3_rotation_matrix(-_a_x)

    # then rotate z to _z by rotating about _x (which is now coincident
    # with x)

    _ra_z = _x
    _a_z = _z.angle(_m_x * z)
    _m_z = _ra_z.axis_and_angle_as_r3_rotation_matrix(-_a_z)

    _m = _m_z * _m_x

    assert math.fabs(_m.determinant() - 1.0) < 1.0e-7

    return _m


def find_undefined_value(cbf_handle):
    """Given a cbf handle, get the value for the undefined pixel."""

    cbf_handle.find_category(b"array_intensities")
    cbf_handle.find_column(b"undefined_value")
    return cbf_handle.get_doublevalue()


def find_underload_value(cbf_handle):
    """Given a cbf handle, get the underload value."""

    cbf_handle.find_category(b"array_intensities")
    cbf_handle.find_column(b"underload")
    return cbf_handle.get_doublevalue()


def find_gain_value(cbf_handle):
    """Given a cbf handle, get the gain value."""
    try:
        cbf_handle.find_category(b"array_intensities")
        cbf_handle.find_column(b"gain")
    except Exception as e:
        if "CBF_NOTFOUND" not in str(e):
            raise
        return 1.0
    return cbf_handle.get_doublevalue()


class detector_helper_sensors:
    """A helper class which allows enumeration of detector sensor technologies
    which should help in identifying specific detectors when needed. These are
    currently limited to IMAGE_PLATE CCD PAD."""

    SENSOR_CCD = "SENSOR_CCD"
    SENSOR_PAD = "SENSOR_PAD"
    SENSOR_IMAGE_PLATE = "SENSOR_IMAGE_PLATE"
    SENSOR_UNKNOWN = "SENSOR_UNKNOWN"

    @staticmethod
    def check_sensor(sensor_type):
        if sensor_type in (
            detector_helper_sensors.SENSOR_CCD,
            detector_helper_sensors.SENSOR_PAD,
            detector_helper_sensors.SENSOR_IMAGE_PLATE,
            detector_helper_sensors.SENSOR_UNKNOWN,
        ):
            return True
        return False

    @staticmethod
    def all():
        return [
            detector_helper_sensors.SENSOR_CCD,
            detector_helper_sensors.SENSOR_PAD,
            detector_helper_sensors.SENSOR_IMAGE_PLATE,
        ]


def set_fast_slow_beam_centre_mm(detector, beam, beam_centre, panel_id=None):
    """detector and beam are dxtbx objects,
    beam_centre is a tuple of (fast, slow) mm coordinates.
    supports 2-theta offset detectors, assumes correct centre provided
    for 2-theta=0
    """
    beam_f, beam_s = beam_centre

    # Ensure panel_id is set
    us0 = matrix.col(beam.get_unit_s0())
    if panel_id is None:
        panel_id = detector.get_panel_intersection(us0)
        if panel_id < 0:
            panel_id = detector.get_panel_intersection(-us0)
            if panel_id < 0:
                panel_id = 0

    # Get data from the chosen panel
    panel = detector[panel_id]
    n = matrix.col(panel.get_normal())

    # Attempt to find the axis an angle of an applied 2theta shift
    cos_angle = n.cos_angle(us0)
    if cos_angle < 0:
        axi = us0.cross(-n)
        ang = us0.angle(-n)
    else:
        axi = us0.cross(n)
        ang = us0.angle(n)

    # Assume a 2theta offset if obliquity >= 5 deg
    two_theta = abs(ang) >= 5.0 * math.pi / 180.0

    # Undo 2theta shift
    if two_theta:
        R = axi.axis_and_angle_as_r3_rotation_matrix(ang)
        Rinv = R.inverse()
        try:
            h = detector.hierarchy()
            h.set_frame(
                fast_axis=Rinv * matrix.col(h.get_fast_axis()),
                slow_axis=Rinv * matrix.col(h.get_slow_axis()),
                origin=Rinv * matrix.col(h.get_origin()),
            )
        except AttributeError:
            for p in detector:
                p.set_frame(
                    fast_axis=Rinv * matrix.col(p.get_fast_axis()),
                    slow_axis=Rinv * matrix.col(p.get_slow_axis()),
                    origin=Rinv * matrix.col(p.get_origin()),
                )

    # Lab coord of desired beam centre
    if us0.accute_angle(n, deg=True) > 89.9:
        raise RuntimeError("Beam is in the plane of the detector panel")
    beam_dist = panel.get_directed_distance() / us0.dot(n)
    beam_centre_lab = beam_dist * us0

    # Lab coord of the current position where we want the beam centre
    intersection_lab = matrix.col(panel.get_lab_coord((beam_f, beam_s)))

    # If the detector has a hierarchy, just update the root note
    try:
        h = detector.hierarchy()
        translation = beam_centre_lab - intersection_lab
        new_origin = matrix.col(h.get_origin()) + translation
        h.set_frame(
            fast_axis=h.get_fast_axis(), slow_axis=h.get_slow_axis(), origin=new_origin
        )
    except AttributeError:
        # No hierarchy, update each panel instead by finding the offset of
        # its origin from the current position of the desired beam centre. Use
        # this to reposition the panel origin wrt the final beam centre
        for p in detector:
            origin = matrix.col(p.get_origin())
            offset = origin - intersection_lab
            new_origin = beam_centre_lab + offset
            p.set_frame(
                fast_axis=p.get_fast_axis(),
                slow_axis=p.get_slow_axis(),
                origin=new_origin,
            )

    # sanity check to make sure we have got the new beam centre correct
    new_beam_centre = detector[panel_id].get_bidirectional_ray_intersection(us0)
    assert (matrix.col(new_beam_centre) - matrix.col((beam_f, beam_s))).length() < 1e-4

    # Re-apply 2theta shift if required
    if two_theta:
        try:
            h = detector.hierarchy()
            h.set_frame(
                fast_axis=R * matrix.col(h.get_fast_axis()),
                slow_axis=R * matrix.col(h.get_slow_axis()),
                origin=R * matrix.col(h.get_origin()),
            )
        except AttributeError:
            for p in detector:
                p.set_frame(
                    fast_axis=R * matrix.col(p.get_fast_axis()),
                    slow_axis=R * matrix.col(p.get_slow_axis()),
                    origin=R * matrix.col(p.get_origin()),
                )

    return


def set_mosflm_beam_centre(detector, beam, mosflm_beam_centre):
    """detector and beam are dxtbx objects,
    mosflm_beam_centre is a tuple of mm coordinates in order (slow, fast).
    supports 2-theta offset detectors, assumes correct centre provided
    for 2-theta=0
    """
    slow_fast_beam_centre = mosflm_beam_centre[1], mosflm_beam_centre[0]
    return set_fast_slow_beam_centre_mm(detector, beam, slow_fast_beam_centre)


def set_detector_distance(detector, distance):
    """
    Set detector origin from distance along normal
    """
    assert len(detector) == 1
    normal = matrix.col(detector[0].get_normal())
    origin = matrix.col(detector[0].get_origin())
    d = origin.dot(normal)
    x = origin - d * normal
    origin = distance * normal + x
    fast_axis = detector[0].get_fast_axis()
    slow_axis = detector[0].get_slow_axis()
    detector[0].set_frame(fast_axis, slow_axis, origin)


def get_detector_projection_2d_axes(
    detector: Detector,
) -> tuple[list[Float2], list[Float2], list[Float2]]:

    """
    Project panel origins, fast and slow axes onto the best-fitting 2D plane.
    """

    # Extract panel vertices
    vertices: list[matrix.rec] = []
    for panel in detector:
        origin = matrix.col(panel.get_origin())
        fast = matrix.col(panel.get_fast_axis())
        slow = matrix.col(panel.get_slow_axis())
        panel_size = panel.get_image_size_mm()
        point1 = origin + panel_size[0] * fast
        point2 = origin + panel_size[1] * slow
        point3 = origin + panel_size[0] * fast + panel_size[1] * slow
        vertices.append(origin)
        vertices.append(point1)
        vertices.append(point2)
        vertices.append(point3)

    # Fit a plane by SVD. Modified from https://stackoverflow.com/a/18968498
    points = np.array(vertices).transpose()
    centre = points.mean(axis=1)
    r = points - centre[:, np.newaxis]  # centroid-to-vertex vectors
    inertia_tensor = np.matmul(r, r.T)
    u = np.linalg.svd(inertia_tensor)[0]
    normal = matrix.col(u[:, 2].tolist()).normalize()
    sum_vertices: matrix.rec = matrix.col((0, 0, 0))
    for v in vertices:
        sum_vertices += v
    if normal.dot(sum_vertices) > 0:
        normal *= -1.0

    # For multi-panel detectors cluster fast, slow axes by DBSCAN to get a
    # consensus X, Y for the 2D plane
    axes: list[matrix.rec]
    clustered_axes = False
    if sklearn and len(detector) > 1:
        clustered_axes = True
        axes = []
        for panel in detector:
            axes.append(matrix.col(panel.get_fast_axis()))
            axes.append(matrix.col(panel.get_slow_axis()))
        clusters = sklearn.cluster.DBSCAN(eps=0.1, min_samples=2).fit_predict(axes)
        nclusters = max(clusters) + 1

        # Revert to single panel mode if clustering is unsuccessful
        if nclusters < 2:
            clustered_axes = False

    if clustered_axes:
        summed_axes: list[matrix.rec] = [
            matrix.col((0, 0, 0)) for i in range(nclusters)
        ]
        for axis, cluster in zip(axes, clusters):
            summed_axes[cluster] += matrix.col(axis)

        # Combine any two clusters approximately related by inversion
        all_checked = False
        while not all_checked:
            all_checked = True
            for i, j in itertools.combinations(range(len(summed_axes)), r=2):
                angle = summed_axes[i].angle(summed_axes[j], deg=True)
                if 175 < angle < 185:
                    summed_axes[i] = summed_axes[i] - summed_axes[j]
                    summed_axes = [e for k, e in enumerate(summed_axes) if k != j]
                    all_checked = False
                    break
        axes = summed_axes
    # For detectors with few panels or badly misaligned panels, align the
    # plane using the corners of the detector
    else:
        # Find the 4 points furthest from the centre (the detector corners)
        dists = np.linalg.norm(r, axis=0)
        threshold = sorted(dists)[-4]
        corners = np.where(dists >= threshold)[0].tolist()
        corners = [vertices[e] for e in corners]

        # Extract two spanning axes from these
        axes = [e - corners[0] for e in corners[1:]]
        lengths = [ax.length() for ax in axes]
        axes = [e for _, e in sorted(zip(lengths, axes), key=itemgetter(0))][0:2]

    # Choose whichever axis is closest to lab X for the plane's X
    labX = matrix.col((1, 0, 0))
    nearX = [ax.accute_angle(labX) for ax in axes]
    X = [ax for _, ax in sorted(zip(nearX, axes))][0]
    if X.dot(labX) < 0:
        X *= -1

    # Choose Y in the plane and X orthogonal
    Y = normal.cross(X).normalize()
    X = Y.cross(normal).normalize()

    # Project centre-shifted origins and fast, slow axes to the plane.
    origin_2d: list[Float2] = []
    fast_2d: list[Float2] = []
    slow_2d: list[Float2] = []
    centre = matrix.col(centre)
    for panel in detector:
        origin = matrix.col(panel.get_origin())
        centre_to_origin = origin - centre
        fast = matrix.col(panel.get_fast_axis())
        slow = matrix.col(panel.get_slow_axis())

        origin_2d.append((centre_to_origin.dot(X), centre_to_origin.dot(Y)))
        fast_2d.append((fast.dot(X), fast.dot(Y)))
        slow_2d.append((slow.dot(X), slow.dot(Y)))

    return origin_2d, fast_2d, slow_2d


def get_panel_projection_2d_from_axes(
    panel: Panel,
    image_data: flex.int | flex.double | flex.float,
    fast_axis_2d: matrix.col,
    slow_axis_2d: matrix.col,
    origin_2d: matrix.col,
) -> tuple[Float4, Float2]:

    """
    Gets translation and rotation required to project image_data from panel,
    based on axes given.
    """

    pixel_size = (
        panel.get_pixel_size()[0] * 1e-3,
        panel.get_pixel_size()[1] * 1e-3,
    )

    center = (
        origin_2d
        + (image_data.focus()[0] - 1) / 2 * pixel_size[1] * slow_axis_2d
        + (image_data.focus()[1] - 1) / 2 * pixel_size[0] * fast_axis_2d
    )
    normal = slow_axis_2d.cross(fast_axis_2d).normalize()

    # Determine rotational and translational components of the
    # homogeneous transformation that maps the readout indices to the
    # three-dimensional laboratory frame.
    Rf = matrix.sqr(
        (
            fast_axis_2d(0, 0),
            fast_axis_2d(1, 0),
            fast_axis_2d(2, 0),
            -slow_axis_2d(0, 0),
            -slow_axis_2d(1, 0),
            -slow_axis_2d(2, 0),
            normal(0, 0),
            normal(1, 0),
            normal(2, 0),
        )
    )
    tf = -Rf * center
    Tf = matrix.sqr(
        (
            Rf(0, 0),
            Rf(0, 1),
            Rf(0, 2),
            tf(0, 0),
            Rf(1, 0),
            Rf(1, 1),
            Rf(1, 2),
            tf(1, 0),
            Rf(2, 0),
            Rf(2, 1),
            Rf(2, 2),
            tf(2, 0),
            0,
            0,
            0,
            1,
        )
    )

    # E maps picture coordinates onto metric Cartesian coordinates,
    # i.e. [row, column, 1 ] -> [x, y, z, 1].  Both frames share the
    # same origin, but the first coordinate of the screen coordinate
    # system increases downwards, while the second increases towards
    # the right.  XXX Is this orthographic projection the only one
    # that makes any sense?
    E = matrix.rec(
        elems=[0, +pixel_size[1], 0, -pixel_size[0], 0, 0, 0, 0, 0, 0, 0, 1],
        n=[4, 3],
    )

    # P: [x, y, z, 1] -> [row, column, 1].  Note that data.focus()
    # needs to be flipped to give (horizontal, vertical) size,
    # i.e. (width, height).
    dim_readout = (image_data.focus()[1], image_data.focus()[0])
    Pf = matrix.rec(
        elems=(
            0,
            -1 / pixel_size[1],
            0,
            (dim_readout[1] - 1) / 2,
            +1 / pixel_size[0],
            0,
            0,
            (dim_readout[0] - 1) / 2,
            0,
            0,
            0,
            1,
        ),
        n=(3, 4),
    )

    # Last row of T is always [0, 0, 0, 1].
    T = Pf * Tf * E
    R = matrix.sqr((T(0, 0), T(0, 1), T(1, 0), T(1, 1)))
    t = matrix.col((T(0, 2), T(1, 2)))

    return cast(Float4, tuple(R)), cast(Float2, tuple(t))



from __future__ import annotations

import os
import pickle

import pytest


@pytest.fixture(scope="session")
def centroid_test_data(dials_regression):
    return os.path.join(dials_regression, "centroid_test_data")


@pytest.fixture
def single_sequence_filenames(centroid_test_data):
    filenames = [
        os.path.join(centroid_test_data, f"centroid_000{i}.cbf") for i in range(1, 10)
    ]
    return filenames


@pytest.fixture
def all_image_examples(dials_regression):
    filenames = (
        ("ALS_1231", "q315r_lyso_1_001.img"),
        ("ALS_501", "als501_q4_1_001.img"),
        ("ALS_821", "q210_lyso_1_101.img"),
        ("ALS_831", "q315r_lyso_001.img"),
        ("APS_14BMC", "q315_1_001.img"),
        ("APS_17ID", "q210_1_001.img"),
        ("APS_19ID", "q315_unbinned_a.0001.img"),
        ("APS_22ID", "mar300.0001"),
        ("APS_23IDD", "mar300_1_E1.0001"),
        ("APS_24IDC", "pilatus_1_0001.cbf"),
        ("APS_24IDC", "q315_1_001.img"),
        ("CLS1_08ID1", "mar225_2_E0_0001.img"),
        ("DESY_ID141", "q210_2_001.img"),
        ("ESRF_BM14", "mar165_001.mccd"),
        ("ESRF_BM14", "mar225_1_001.mccd"),
        ("ESRF_ID231", "q315r_7_001.img"),
        ("RAXIS-HTC", "test1_lysozyme_0111060001.osc"),
        ("SLS_X06SA", "mar225_2_001.img"),
        ("SLS_X06SA", "pilatus6m_1_00001.cbf"),
        ("SRS_101", "mar225_001.img"),
        ("SRS_142", "q4_1_001.img"),
        ("SSRL_bl111", "mar325_1_001.mccd"),
        ("xia2", "merge2cbf_averaged_0001.cbf"),
    )
    return [os.path.join(dials_regression, "image_examples", *f) for f in filenames]


@pytest.fixture
def multiple_block_filenames(single_sequence_filenames, all_image_examples):
    return single_sequence_filenames + all_image_examples


def pickle_then_unpickle(obj):
    return pickle.loads(pickle.dumps(obj))



"""
A top-level class to represent image formats which does little else but
(i) establish an abstract class for what needs to be implemented and
(ii) include the format registration code for any image formats which
inherit from this. This will also contain links to the static methods
from the X(component)Factories which will allow construction of e.g.
goniometers etc. from the headers and hence a format specific factory.
"""

from __future__ import annotations

import bz2
import functools
import os
from io import IOBase
from typing import Callable, ClassVar

import libtbx

import dxtbx.filecache_controller
from dxtbx.format.image import ImageBool
from dxtbx.model import MultiAxisGoniometer
from dxtbx.model.beam import BeamFactory
from dxtbx.model.detector import DetectorFactory
from dxtbx.model.goniometer import GoniometerFactory
from dxtbx.model.scan import ScanFactory
from dxtbx.sequence_filenames import template_regex
from dxtbx.util import get_url_scheme

try:
    import gzip
except ImportError:
    gzip = None  # type: ignore


_cache_controller = dxtbx.filecache_controller.simple_controller()


def abstract(cls):
    """
    Mark a Format class as abstract, but not it's subclasses.

    This means that it should not be directly used to parse user images.
    """

    cls._abstract_base = cls
    return cls


class Reader:
    def __init__(self, format_class, filenames, **kwargs):
        self._kwargs = kwargs
        self.format_class = format_class
        self._filenames = filenames

    def read(self, index):
        format_instance = self.format_class.get_instance(
            self._filenames[index], **self._kwargs
        )
        return format_instance.get_raw_data()

    def paths(self):
        return self._filenames

    def identifiers(self):
        return self._filenames

    def __len__(self):
        return len(self._filenames)

    def copy(self, filenames):
        return Reader(self.format_class, filenames, **self._kwargs)

    def is_single_file_reader(self):
        return False

    def master_path(self):
        return ""


@abstract
class Format:
    """A base class for the representation and interrogation of diffraction
    image formats, from which all classes for reading the header should be
    inherited. This includes: autoregistration of implementation classes,
    stubs which need to be overridden and links to static factory methods
    which will prove to be useful in other implementations.

    Attributes:
        schemes: List of schemes this Format class supports.
            An empty string ("") is treated as non-URL filenames, where
            "file" is treated as explicit file:// support. Format classes
            that do not support the listed schema will _not_ be called
            to understand() with the locator - note that this means if a
            class does not declare "file", it will not get called to
            understand "file://" URIs passed to it. An empty `schemes` will
            not ever get called in Format registry searches.
    """

    schemes: list[str] = [""]

    # Which class is the abstract base. Assigned with the @abstract
    # decorator, and used to check initialization
    _abstract_base: ClassVar[Format | None] = None

    @staticmethod
    def understand(image_file):
        """Overload this to publish whether this class instance
        understands a given file.  N.B. to say that we understand it,
        return True.  If a subclass also understands the image
        (because, for example, its detector serial number takes a
        certain value), it will by definition understand it better
        than its superclass.  Thus, the preferred class will be the
        deepest subclass in the inheritance hierarchy.  Finally, if
        you are writing this subclass for a given instrument and you
        are given a different example return False.

        Implementing understand() in a subclass, one can safely assume
        that the superclasss understand() function returned True.
        The understand() function of two different classes directly
        derived from the same base should never both return True for
        the same input image."""

        return False

    @classmethod
    def is_abstract(cls) -> bool:
        """
        Is this Format class allowed to be instantiated?

        Format classes that describe an abstract class of formats but do
        not explicitly read a file-format should be marked as such with
        the dxtbx.format.abstract decorator, and they will not be
        instantiable or used as a registry match via load lookups.
        """
        return cls is cls._abstract_base

    @staticmethod
    def has_dynamic_shadowing(**kwargs):
        dynamic_shadowing = kwargs.get("dynamic_shadowing", False)
        if dynamic_shadowing in (libtbx.Auto, "Auto"):
            return False
        return dynamic_shadowing

    def __init__(self, image_file, **kwargs):
        """Initialize a class instance from an image file."""

        # Don't allow abstract instantion
        # - except for Format, which is used as a placeholder in many
        # places (e.g. still, check_format=False) so needs to be allowed.
        if self.is_abstract() and not type(self) is Format:
            raise TypeError(
                f"Cannot instantiate: Format class '{type(self).__name__}' is marked abstract"
            )

        self._image_file = image_file

        self._goniometer_instance = None
        self._detector_instance = None
        self._beam_instance = None
        self._scan_instance = None

        self._goniometer_factory = GoniometerFactory
        self._detector_factory = DetectorFactory
        self._beam_factory = BeamFactory
        self._scan_factory = ScanFactory

        self.setup()

    def setup(self):
        """Read the image file, construct the information which we will be
        wanting about the experiment from this. N.B. in your implementation
        of this you will probably want to make use of the static methods
        below and probably add some format parsing code too. Please also keep
        in mind that your implementation may be further subclassed by
        someone else."""

        self._start()

        try:
            goniometer_instance = self._goniometer()
            self._goniometer_instance = goniometer_instance

            detector_instance = self._detector()
            self._detector_instance = detector_instance

            beam_instance = self._beam()
            self._beam_instance = beam_instance

            scan_instance = self._scan()
            self._scan_instance = scan_instance
        finally:
            self._end()

    @staticmethod
    def get_cache_controller():
        return _cache_controller

    def get_goniometer(self):
        """Get the standard goniometer instance which was derived from the
        image headers."""

        return self._goniometer_instance

    def get_detector(self):
        """Get the standard detector instance which was derived from the
        image headers."""

        return self._detector_instance

    def get_beam(self):
        """Get the standard beam instance which was derived from the image
        headers."""

        return self._beam_instance

    def get_scan(self):
        """Get the standard scan instance which was derived from the image
        headers."""

        return self._scan_instance

    def get_raw_data(self):
        """Get the pixel intensities (i.e. read the image and return as a
        flex array."""
        try:
            image = self.detectorbase
        except AttributeError:
            return None
        image.read()
        raw_data = image.get_raw_data()
        return raw_data

    def get_vendortype(self):
        return "no dxtbx Format vendortype"

    def detectorbase_start(self):
        raise NotImplementedError

    def get_detectorbase(self):
        """Return the instance of detector base."""
        self.detectorbase_start()

        # Proxy to link dxtbx to iotbx
        class _detectorbase_proxy:
            def __init__(self, format_instance):
                self._fi = format_instance
                if not hasattr(self, "vendortype"):
                    self.vendortype = "generic"

            def __getattribute__(self, name):
                if name == "__class__":
                    return self._fi.__class__
                return object.__getattribute__(self, name)

            def __getattr__(self, name):
                try:
                    return self._fi.__getattribute__(name)
                except AttributeError:
                    try:
                        return self._fi.detectorbase.__getattribute__(name)
                    except AttributeError:
                        return self._fi.detectorbase.__getattr__(name)

        return _detectorbase_proxy(self)

    @classmethod
    def get_instance(Class, filename, **kwargs):
        if (
            not hasattr(Class, "_current_instance_")
            or Class._current_filename_ != filename
            or Class._current_kwargs_ != kwargs
        ):
            Class._current_instance_ = Class(filename, **kwargs)
            Class._current_filename_ = filename
            Class._current_kwargs_ = kwargs
        return Class._current_instance_

    @classmethod
    def get_reader(cls):
        """
        Return a reader class
        """
        return functools.partial(Reader, cls)

    def get_masker(self, goniometer=None):
        """
        Return a masker class
        """
        if (
            isinstance(goniometer, MultiAxisGoniometer)
            and hasattr(self, "_dynamic_shadowing")
            and self._dynamic_shadowing
        ):
            masker = self.get_goniometer_shadow_masker(goniometer=goniometer)
        else:
            masker = None
        return masker

    @classmethod
    def get_imageset(
        Class,
        input_filenames,
        beam=None,
        detector=None,
        goniometer=None,
        scan=None,
        as_imageset=False,
        as_sequence=False,
        single_file_indices=None,
        format_kwargs=None,
        template=None,
        check_format=True,
    ):
        """
        Factory method to create an imageset

        """
        # Import here to avoid cyclic imports
        from dxtbx.imageset import ImageSequence, ImageSet, ImageSetData

        # Turn entries that are filenames into absolute paths
        filenames = [
            os.fspath(os.path.abspath(x)) if not get_url_scheme(x) else x
            for x in input_filenames
        ]

        # Make it a dict
        if format_kwargs is None:
            format_kwargs = {}

        # Get some information from the format class
        reader = Class.get_reader()(filenames, **format_kwargs)

        # Get the format instance
        if check_format is True:
            Class._current_filename_ = None
            Class._current_instance_ = None
            format_instance = Class.get_instance(filenames[0], **format_kwargs)
        else:
            format_instance = None

        # Read the vendor type
        if check_format is True:
            vendor = format_instance.get_vendortype()
        else:
            vendor = ""

        # Get the format kwargs
        params = format_kwargs

        # Make sure only 1 or none is set
        assert [as_imageset, as_sequence].count(True) < 2
        if as_imageset:
            is_sequence = False
        elif as_sequence:
            is_sequence = True
        else:
            if scan is None and format_instance is None:
                raise RuntimeError(
                    """
          One of the following needs to be set
            - as_imageset=True
            - as_sequence=True
            - scan
            - check_format=True
      """
                )
            if scan is None:
                test_scan = format_instance.get_scan()
            else:
                test_scan = scan
            if test_scan is not None and not test_scan.is_still():
                is_sequence = True
            else:
                is_sequence = False

        # Create an imageset or sequence
        if not is_sequence:

            # Create the imageset
            iset = ImageSet(
                ImageSetData(
                    reader=reader,
                    masker=None,
                    vendor=vendor,
                    params=params,
                    format=Class,
                )
            )

            # If any are None then read from format
            if [beam, detector, goniometer, scan].count(None) != 0:

                # Get list of models
                beam = []
                detector = []
                goniometer = []
                scan = []
                for f in filenames:
                    format_instance = Class(f, **format_kwargs)
                    beam.append(format_instance.get_beam())
                    detector.append(format_instance.get_detector())
                    goniometer.append(format_instance.get_goniometer())
                    scan.append(format_instance.get_scan())

            # Set the list of models
            for i in range(len(filenames)):
                iset.set_beam(beam[i], i)
                iset.set_detector(detector[i], i)
                iset.set_goniometer(goniometer[i], i)
                iset.set_scan(scan[i], i)

        else:

            # Get the template
            if template is None:
                template = template_regex(filenames[0])[0]
            else:
                template = str(template)

            # Check scan makes sense
            if scan:
                if check_format is True:
                    assert scan.get_num_images() == len(filenames)

            # If any are None then read from format
            if beam is None and format_instance is not None:
                beam = format_instance.get_beam()
            if detector is None and format_instance is not None:
                detector = format_instance.get_detector()
            if goniometer is None and format_instance is not None:
                goniometer = format_instance.get_goniometer()
            if scan is None and format_instance is not None:
                scan = format_instance.get_scan()
                if scan is not None:
                    for f in filenames[1:]:
                        format_instance = Class(f, **format_kwargs)
                        scan += format_instance.get_scan()

            assert beam is not None, "Can't create Sequence without beam"
            assert detector is not None, "Can't create Sequence without detector"
            assert goniometer is not None, "Can't create Sequence without goniometer"
            assert scan is not None, "Can't create Sequence without scan"

            # Create the masker
            if format_instance is not None:
                masker = format_instance.get_masker(goniometer=goniometer)
            else:
                masker = None

            # Create the sequence
            iset = ImageSequence(
                ImageSetData(
                    reader=reader,
                    masker=masker,
                    vendor=vendor,
                    params=params,
                    format=Class,
                    template=template,
                ),
                beam=beam,
                detector=detector,
                goniometer=goniometer,
                scan=scan,
            )

        if format_instance is not None:
            static_mask = format_instance.get_static_mask()
            if static_mask is not None:
                if not iset.external_lookup.mask.data.empty():
                    for m1, m2 in zip(static_mask, iset.external_lookup.mask.data):
                        m1 &= m2.data()
                    iset.external_lookup.mask.data = ImageBool(static_mask)
                else:
                    iset.external_lookup.mask.data = ImageBool(static_mask)

        return iset

    def get_image_file(self):
        """Get the image file provided to the constructor."""

        return self._image_file

    # methods which must be overloaded in order to produce a useful Format
    # class implementation

    def _start(self):
        """Start code for handling this image file, which may open a link
        to it once, say, and pass this around within the implementation.
        How you use this is up to you, though you probably want to overload
        it..."""

        return

    def _end(self):
        """Clean up things - keeping in mind that this should run even in the
        case of an exception being raised."""

        return

    def _goniometer(self):
        """Overload this method to read the image file however you like so
        long as the result is a goniometer."""
        return None

    def _detector(self):
        """Overload this method to read the image file however you like so
        long as the result is a detector."""
        return None

    def _beam(self):
        """Overload this method to read the image file however you like so
        long as the result is a beam."""
        return None

    def _scan(self):
        """Overload this method to read the image file however you like so
        long as the result is a scan."""
        return None

    def get_static_mask(self):
        """Overload this method to override the static mask."""
        return None

    def get_spectrum(self):
        """Overload this method to read the image file however you like so
        long as the result is a spectrum
        """
        return None

    def get_goniometer_shadow_masker(self, goniometer=None):
        """Overload this method to allow generation of dynamic goniometer shadow
        masks to be used during spotfinding or integration."""

        return None

    ####################################################################
    #                                                                  #
    # Helper functions for dealing with compressed images.             #
    #                                                                  #
    ####################################################################

    @classmethod
    def open_file(cls, filename: str | bytes | os.PathLike, mode="rb"):
        """Open file for reading, decompressing silently if necessary,
        caching transparently if possible."""
        filename_str = os.fspath(filename)
        assert isinstance(filename_str, str)

        fh_func: Callable[..., IOBase]
        if filename_str.endswith(".bz2"):
            fh_func = functools.partial(bz2.BZ2File, filename, mode=mode)

        elif filename_str.endswith(".gz"):
            if not gzip:
                raise RuntimeError("gz file provided without gzip module")
            fh_func = functools.partial(gzip.GzipFile, filename, mode=mode)

        else:
            fh_func = functools.partial(open, filename, mode=mode)

        ##  To disable caching logic:
        # return fh_func()
        return cls.get_cache_controller().check(filename_str, fh_func)



from __future__ import annotations

import sys

import h5py
import numpy as np

from iotbx.detectors.eiger import EIGERImage
from scitbx import matrix

from dxtbx.format.FormatHDF5 import FormatHDF5
from dxtbx.format.FormatPilatusHelpers import determine_eiger_mask
from dxtbx.format.FormatPilatusHelpers import get_vendortype_eiger as gv
from dxtbx.format.nexus import (
    BeamFactory,
    DataFactory,
    DataFactoryCache,
    DetectorFactory,
    GoniometerFactory,
    MaskFactory,
    NXdata,
    NXmxReader,
    generate_scan_model,
)
from dxtbx.util import ersatz_uuid4


def find_entries(nx_file):
    """
    Find NXmx entries
    """
    if "entry" in nx_file:
        entry = nx_file["entry"]
        if "NX_class" in entry.attrs:
            if entry.attrs["NX_class"] == np.string_("NXentry"):
                if "definition" not in entry:
                    return entry
    return None


def is_eiger_nearly_nexus_file(filename):
    """
    A hacky function to check if this is an EIGER-flavoured nexus file
    """
    # Get the file handle
    with h5py.File(filename, "r") as handle:
        # Find the NXmx entries
        entry = find_entries(handle)
        if entry is not None:
            try:
                return (
                    np.string_("dectris eiger")
                    in entry["instrument"]["detector"]["description"][()].lower()
                )
            except KeyError:
                pass
        return False


class EigerNXmxFixer:
    """
    A hacky class to read an NXmx file
    """

    def __init__(self, input_filename, memory_mapped_name):
        # Copy the master file to the in memory handle
        handle_orig = h5py.File(input_filename, "r")
        handle = h5py.File(
            name=memory_mapped_name, driver="core", backing_store=False, mode="w"
        )
        handle_orig.copy("entry", handle)

        # Add some simple datasets
        def create_scalar(handle, path, dtype, value):
            dataset = handle.create_dataset(path, (), dtype=dtype)
            dataset[()] = value

        # Add NXmx definition
        create_scalar(handle["entry"], "definition", "S4", np.string_("NXmx"))

        # Add saturation value
        try:
            create_scalar(
                handle["entry/instrument/detector"],
                "saturation_value",
                "int32",
                handle[
                    "/entry/instrument/detector/detectorSpecific/countrate_correction_count_cutoff"
                ],
            )
        except Exception:
            create_scalar(
                handle["entry/instrument/detector"],
                "saturation_value",
                "int32",
                handle[
                    "entry/instrument/detector/detectorSpecific/detectorModule_000/countrate_correction_count_cutoff"
                ],
            )

        # Add detector type
        create_scalar(
            handle["entry/instrument/detector"], "type", "S5", np.string_("PIXEL")
        )

        # Move the beam
        # print "Copying /entry/instrument/beam to /entry/sample/beam"
        handle.copy("/entry/instrument/beam", "/entry/sample/beam")

        # Create detector module
        module_path = "/entry/instrument/detector/module"
        # print "Creating detector module %s" % (module_path)
        group = handle.create_group(module_path)
        group.attrs["NX_class"] = np.string_("NXdetector_module")

        # Add a module index
        create_scalar(group, "module_index", "int64", 0)

        # Create detector data origin
        dataset = group.create_dataset("data_origin", (2,), dtype="int32")
        dataset[0] = 0
        dataset[1] = 0

        # cope with badly structured chunk information i.e. many more data
        # entries than there are in real life...
        delete = []
        handle_orig_entry_properties = {}
        self.data_factory_cache = {}
        for k in sorted(handle_orig["/entry/data"]):
            try:
                handle_orig_entry = handle_orig["/entry/data/%s" % k]
                shape = handle_orig_entry.shape
            except KeyError:
                delete.append("/entry/data/%s" % k)
                continue
            handle_orig_entry_properties[k] = {
                "shape": shape,
                "length": len(handle_orig_entry),
                "filename": handle_orig_entry.file.filename,
            }
            self.data_factory_cache[k] = DataFactoryCache(
                shape=shape,
                ndim=handle_orig_entry.ndim,
                filename=handle_orig_entry.file.filename,
                is_virtual=False,
            )
        for d in delete:
            del handle[d]

        # Create detector data size
        dataset = group.create_dataset("data_size", (2,), dtype="int32")
        dataset[0] = handle_orig_entry_properties["data_000001"]["shape"][2]
        dataset[1] = handle_orig_entry_properties["data_000001"]["shape"][1]

        depends_on = "/entry/instrument/detector/transformations/translation"

        # Add fast_pixel_size dataset
        # print "Using /entry/instrument/detector/geometry/orientation/value as fast/slow pixel directions"
        fast_axis = handle["/entry/instrument/detector/geometry/orientation/value"][0:3]
        fast_axis = [
            fast_axis[0],
            fast_axis[1],
            -fast_axis[2],
        ]  # swap Z axis to align with Dectris/NeXus documentation
        slow_axis = handle["/entry/instrument/detector/geometry/orientation/value"][3:6]
        slow_axis = [
            slow_axis[0],
            slow_axis[1],
            -slow_axis[2],
        ]  # swap Z axis to align with Dectris/NeXus documentation
        create_scalar(
            group,
            "fast_pixel_direction",
            "float32",
            handle["/entry/instrument/detector/x_pixel_size"][()],
        )
        group["fast_pixel_direction"].attrs["transformation_type"] = np.string_(
            "translation"
        )
        group["fast_pixel_direction"].attrs["vector"] = fast_axis
        group["fast_pixel_direction"].attrs["offset"] = (0, 0, 0)
        group["fast_pixel_direction"].attrs["units"] = np.string_("m")
        group["fast_pixel_direction"].attrs["depends_on"] = np.string_(depends_on)

        # Add slow_pixel_size dataset
        create_scalar(
            group,
            "slow_pixel_direction",
            "float32",
            handle["/entry/instrument/detector/y_pixel_size"][()],
        )
        group["slow_pixel_direction"].attrs["transformation_type"] = np.string_(
            "translation"
        )
        group["slow_pixel_direction"].attrs["vector"] = slow_axis
        group["slow_pixel_direction"].attrs["offset"] = (0, 0, 0)
        group["slow_pixel_direction"].attrs["units"] = np.string_("m")
        group["slow_pixel_direction"].attrs["depends_on"] = np.string_(depends_on)

        # Add module offset dataset
        # print "Set module offset to be zero relative to detector"
        create_scalar(group, "module_offset", "float32", 0)
        group["module_offset"].attrs["transformation_type"] = np.string_("translation")
        group["module_offset"].attrs["vector"] = (0, 0, 0)
        group["module_offset"].attrs["offset"] = (0, 0, 0)
        group["module_offset"].attrs["units"] = np.string_("m")
        group["module_offset"].attrs["depends_on"] = np.string_(depends_on)

        # Create detector depends_on
        create_scalar(
            handle["/entry/instrument/detector"],
            "depends_on",
            "S%d" % len(depends_on),
            np.string_(depends_on),
        )

        # Add detector position
        # print "Using /entry/instrument/detector/geometry/translation/distances as transformation"
        detector_offset_vector = handle[
            "/entry/instrument/detector/geometry/translation/distances"
        ][()]
        # swap Z axis to align with Dectris/NeXus documentation
        detector_offset_vector = matrix.col(
            (
                detector_offset_vector[0],
                detector_offset_vector[1],
                -detector_offset_vector[2],
            )
        )
        group = handle.create_group("/entry/instrument/detector/transformations")
        group.attrs["NX_class"] = np.string_("NXtransformations")
        create_scalar(group, "translation", "float32", detector_offset_vector.length())
        group["translation"].attrs["transformation_type"] = np.string_("translation")
        if detector_offset_vector.length() > 0:
            group["translation"].attrs["vector"] = detector_offset_vector.normalize()
        else:
            group["translation"].attrs["vector"] = detector_offset_vector
        group["translation"].attrs["offset"] = 0
        group["translation"].attrs["units"] = np.string_("m")
        group["translation"].attrs["depends_on"] = np.string_(".")

        # Create goniometer transformations if not found
        if "/entry/sample/transformations" not in handle:
            # print "Creating group /entry/sample/transformation"
            group = handle.create_group("/entry/sample/transformations")
            group.attrs["NX_class"] = np.string_("NXtransformations")
        else:
            group = handle["/entry/sample/transformations"]

        # check for incomplete omega definitions dirty hack...
        if "omega" in group:
            try:
                group["omega"][()]
            except (AttributeError, TypeError):
                del group["omega"]

        if "omega" not in group:
            # In here assume goniometer axis is 1,0,0 unless (i) specified somewhere
            # we can know or (ii) a known special case. For (i) for this instrument
            # listed here this is at
            #
            # /entry/sample/transformations/omega->vector
            #
            # which will join the special case list once this is properly resolved
            # as this corrently returns 0, -1, 0 so needs transforming...
            #
            # special cases:
            # E-32-0105 - Max IV, vertical axis

            key = handle["/entry/instrument/detector/detector_number"][()]
            default_axis = {b"E-32-0105": (0, 1, 0)}.get(key, (-1, 0, 0))

            num_images = 0
            for name in sorted(handle["/entry/data"]):
                num_images += handle_orig_entry_properties[name]["length"]
            dataset = group.create_dataset("omega", (num_images,), dtype="float32")
            dataset.attrs["units"] = np.string_("degree")
            dataset.attrs["transformation_type"] = np.string_("rotation")
            dataset.attrs["vector"] = default_axis
            dataset.attrs["offset"] = 0
            dataset.attrs["depends_on"] = np.string_(".")
            omega_range_average = handle[
                "/entry/sample/goniometer/omega_range_average"
            ][()]
            omega_range_average = int(omega_range_average * 100 + 0.5) / 100.0
            for i in range(num_images):
                angle = omega_range_average * i
                dataset[i] = angle
        else:
            dataset = group["omega"]

        if "depends_on" not in handle["/entry/sample"]:
            # Create sample depends_on
            create_scalar(
                handle["/entry/sample"],
                "depends_on",
                "S%d" % len(dataset.name),
                np.string_(dataset.name),
            )

        # Change relative paths to absolute paths
        for name in handle["/entry/data"]:
            del handle["entry/data"][name]
            filename = handle_orig_entry_properties[name]["filename"]
            handle["entry/data"][name] = h5py.ExternalLink(filename, "entry/data/data")
            handle["entry/data"]["_filename_" + name] = filename  # Store file names

        self.handle = handle
        self.handle_orig = handle_orig


class FormatHDF5EigerNearlyNexus(FormatHDF5):
    @staticmethod
    def understand(image_file):
        try:
            return is_eiger_nearly_nexus_file(image_file)
        except OSError:
            return False

    def _start(self):
        # Read the file structure
        temp_file = "tmp_master_%s.nxs" % ersatz_uuid4()
        fixer = EigerNXmxFixer(self._image_file, temp_file)
        reader = NXmxReader(handle=fixer.handle)

        # Only support 1 set of models at the moment
        assert len(reader.entries) == 1, "Currently only supports 1 NXmx entry"
        assert len(reader.entries[0].data) == 1, "Currently only supports 1 NXdata"
        assert (
            len(reader.entries[0].instruments) == 1
        ), "Currently only supports 1 NXinstrument"
        assert len(reader.entries[0].samples) == 1, "Currently only supports 1 NXsample"
        assert (
            len(reader.entries[0].instruments[0].detectors) == 1
        ), "Currently only supports 1 NXdetector"
        assert (
            len(reader.entries[0].instruments[0].detectors[0].modules) == 1
        ), "Currently only supports 1 NXdetector_module"
        assert (
            len(reader.entries[0].samples[0].beams) == 1
            or len(reader.entries[0].instruments[0].beams) == 1
        ), "Currently only supports 1 NXbeam"

        # Get the NXmx model objects
        entry = reader.entries[0]
        self.instrument = instrument = entry.instruments[0]
        detector = instrument.detectors[0]
        sample = entry.samples[0]
        beam = sample.beams[0] if sample.beams else instrument.beams[0]

        # Use data from original master file
        data = NXdata(fixer.handle_orig[entry.data[0].handle.name])
        self._raw_data = DataFactory(data, cached_information=fixer.data_factory_cache)

        # Construct the models
        self._beam_factory = BeamFactory(beam)
        self._beam_factory.load_model(0)
        self._detector_model = DetectorFactory(
            detector, self._beam_factory.model, shape=self._raw_data.shape()
        ).model

        # Override the minimum trusted value - for Eiger should be 0.
        # This needs a careful check in the context of testing the
        # trust-the-trusted_range branch
        for panel in self._detector_model:
            trusted = panel.get_trusted_range()
            panel.set_trusted_range((0, trusted[1]))

        self._goniometer_model = GoniometerFactory(sample).model
        self._scan_model = generate_scan_model(sample, detector)

        # update model for masking Eiger detectors
        for f0, f1, s0, s1 in determine_eiger_mask(self._detector_model):
            self._detector_model[0].add_mask(f0 - 1, s0 - 1, f1, s1)

    def _end(self):
        return

    def _goniometer(self):
        return self._goniometer_model

    def _detector(self):
        return self._detector_model

    def _beam(self, index=None):
        self._beam_model, _ = self._beam_factory.read_models(index)
        return self._beam_model

    def _scan(self):
        return self._scan_model

    def get_goniometer(self, index=None):
        return self._goniometer()

    def get_detector(self, index=None):
        return self._detector()

    def get_beam(self, index=None):
        return self._beam(index)

    def get_scan(self, index=None):
        if index is None:
            return self._scan()
        return self._scan()[index]

    def get_raw_data(self, index):
        return self._raw_data[index]

    def get_static_mask(self, index=None, goniometer=None):
        return MaskFactory(self.instrument.detectors, index).mask

    def get_num_images(self):
        scan = self._scan()
        if isinstance(scan, list):
            return sum(s.get_num_images() for s in scan)
        return scan.get_num_images()

    def get_image_file(self, index=None):
        return self._image_file

    def detectorbase_start(self, index=0):
        self.detectorbase = EIGERImage(self._image_file, index=index)
        self.detectorbase.readHeader(dxtbx_instance=self)

        def model_get_raw_data(ptr, index):
            return self.get_raw_data(index)

        self.detectorbase.get_raw_data_callback = model_get_raw_data

    def get_detectorbase(self, index=0):
        self.detectorbase_start(index)
        return self.detectorbase

    def get_vendortype(self):
        return gv(self.get_detector())


if __name__ == "__main__":
    f = FormatHDF5EigerNearlyNexus(sys.argv[1])
    for i in range(10):
        print(f.get_raw_data(i))



"""
Subclass of FormatTiffRayonix to deal with images who have beam centers
specified in pixels
"""

from __future__ import annotations

import struct
import sys

from scitbx.matrix import col

from dxtbx.format.FormatTIFFHelpers import BIG_ENDIAN, LITTLE_ENDIAN
from dxtbx.format.FormatTIFFRayonix import FormatTIFFRayonix


class FormatTIFFRayonixESRF(FormatTIFFRayonix):
    """A class for reading TIFF format Rayonix images, and correctly
    constructing a model for the experiment from this."""

    @staticmethod
    def understand(image_file):
        """Check to see if this looks like an Rayonix TIFF format image,
        i.e. we can make sense of it.  Returns true if the beam center is specified
        in pixels."""

        width, height, depth, order, bytes = FormatTIFFRayonix.get_tiff_header(
            image_file
        )

        format = {LITTLE_ENDIAN: "<", BIG_ENDIAN: ">"}[order]
        offset = 1024

        detector_size_pixels = col(
            struct.unpack(format + "ii", bytes[offset + 80 : offset + 88])
        )
        detector_center_px = 0.5 * detector_size_pixels

        # detector_pixel_sz_mm = 1.0e-6 * col(  # convert from nano to milli
        #     struct.unpack(format + "ii", bytes[offset + 772 : offset + 780])
        # )

        header_beam_center = 0.001 * col(  # Rayonix says this should be pixels
            struct.unpack(format + "ii", bytes[offset + 644 : offset + 652])
        )

        disagreement = header_beam_center[0] / detector_center_px[0]
        return (
            header_beam_center[0] > 0
            and header_beam_center[1] > 0
            and disagreement < 0.5
        )  # if header was in mm, disagreement should be
        # approximately the pixel size in mm

    def _goniometer(self):
        """Return a model for goniometer corresponding to the values stored
        in the image header. In the first instance assume this is a single
        axis annd raise exception otherwise."""

        # starts, ends, offset, width = self._get_rayonix_scan_angles()

        # not testing as this the CLS images are not properly structured...
        # and also don't have a serial number in (FAIL)

        return self._goniometer_factory.single_axis()

    ####################################################################
    #                                                                  #
    # Helper methods to get all of the values out of the TIFF header   #
    # - separated out to assist with code clarity                      #
    #                                                                  #
    ####################################################################

    def _get_rayonix_beam_xy(self):
        """Get the beam x, y positions which are defined in the standard
        to be in pixels. X and Y are not defined by the documentation, so
        taking as a given that these are horizontal and vertical. N.B.
        the documentation states that the horizontal direction is fast."""

        beam_x, beam_y = struct.unpack(self._ii, self._tiff_header_bytes[1668:1676])[:2]
        pixel_x, pixel_y = struct.unpack(self._ii, self._tiff_header_bytes[1796:1804])[
            :2
        ]

        return beam_x * 1000 / pixel_x, beam_y * 1000 / pixel_y


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatTIFFRayonixESRF.understand(arg))



from __future__ import annotations

import json
import os

from dxtbx.model.crystal import CrystalFactory
from dxtbx.serialize.imageset import imageset_from_dict


def imageset(filename):
    """Load the given JSON file.

    Params:
        infile The input filename

    Returns:
        The models

    """
    # If the input is a string then open and read from that file
    filename = os.path.abspath(filename)
    directory = os.path.dirname(filename)
    with open(filename) as infile:
        return imageset_from_dict(json.load(infile), directory=directory)


def crystal(infile):
    """Load the given JSON file.

    Params:
        infile The input filename or file object

    Returns:
        The models

    """
    # If the input is a string then open and read from that file
    if isinstance(infile, str):
        with open(infile) as infile:
            return CrystalFactory.from_dict(json.loads(infile.read()))

    # Otherwise assume the input is a file and read from it
    else:
        return CrystalFactory.from_dict(json.loads(infile.read()))


def experiment_list(infile, check_format=True):
    """Load an experiment list from a serialized format."""
    # Resolve recursive import
    from dxtbx.model.experiment_list import ExperimentListFactory

    if infile and hasattr(infile, "__fspath__"):
        infile = (
            infile.__fspath__()
        )  # Resolve file system path (PEP-519) object to string.

    return ExperimentListFactory.from_serialized_format(
        infile, check_format=check_format
    )



from __future__ import annotations

from unittest.mock import Mock, create_autospec

import pytest

import dxtbx.filecache
import dxtbx.filecache_controller as fcc


def test_invalid_cache(monkeypatch):
    """Tests a condition where the filecache controller can give out an invalid cache"""

    mocklazy = create_autospec(dxtbx.filecache.lazy_file_cache)
    monkeypatch.setattr(fcc.dxtbx.filecache, "lazy_file_cache", mocklazy)

    # Create the cache
    cache = fcc.simple_controller()

    # Set up an initial, working cache
    good_file_opener = Mock()
    cache.check("working", lambda: good_file_opener)
    # This should have used the lazy_file_cache to set up
    mocklazy.assert_called_with(good_file_opener)
    mocklazy.return_value.open.assert_called()

    # Now, pass the cache an opener that fails
    badfile = Mock(side_effect=IOError("Testing bad file"))
    with pytest.raises(IOError):
        cache.check("not_working", badfile)

    # The bug: Calling check with the same tag shouldn't use the invalid cache
    # To Test: Do the working test, but with the failed tag. If the invalid
    #          cache is used, then we won't have asked for reconstruction
    mocklazy.reset_mock()
    cache.check("not_working", lambda: good_file_opener)
    mocklazy.assert_called_with(good_file_opener)
    mocklazy.return_value.open.assert_called()



"""An implementation of the CBF image reader for Eiger images"""


from __future__ import annotations

import os
import sys

from iotbx.detectors.pilatus_minicbf import PilatusImage

from dxtbx.format.FormatCBFMini import FormatCBFMini
from dxtbx.format.FormatPilatusHelpers import get_vendortype_eiger as gv


class FormatCBFMiniEigerPhotonFactory(FormatCBFMini):
    """A class for reading mini CBF format Eiger images, and correctly
    constructing a model for the experiment from this.

    Specific for 2-Eiger setup at Photon Factory based on images sent by
    Yusuke Yamada. Geometry currently hard-coded as no geometry information in
    image headers. Assumes image filenames contain the string '_upper_' or
    '_lower_' to distinguish the two detectors.

    Only works if environment variable ENABLE_PHOTON_FACTORY_TWO_EIGER is set.
    """

    @staticmethod
    def understand(image_file):
        # no way to uniquely identify given example images

        if "ENABLE_PHOTON_FACTORY_TWO_EIGER" in os.environ:
            return True
        return False

    def _goniometer(self):
        return self._goniometer_factory.make_goniometer(
            (1, 0, 0), (1, 0, 0, 0, 1, 0, 0, 0, 1)
        )

    def _detector(self):

        max_trusted_value = 1e7

        if "_lower_" in self._image_file:
            # Detector:
            # Panel:
            # pixel_size:{0.075,0.075}
            # image_size: {2070,2167}
            # trusted_range: {0,1e+07}
            # thickness: 0
            # material:
            # mu: 0
            # fast_axis: {0.999997,-0.00228845,0.00127629}
            # slow_axis: {-0.0026197,-0.862841,0.505469}
            # origin: {-79.3767,2.47119,-169.509}

            detector = self._detector_factory.complex(
                "PAD",
                origin=(-79.3767, 2.47119, -169.509),
                fast=(0.999997, -0.00228845, 0.00127629),
                slow=(-0.0026197, -0.862841, 0.505469),
                pixel=(0.075, 0.075),
                size=(2070, 2167),
                trusted_range=(0, max_trusted_value),
            )

        elif "_upper_" in self._image_file:
            # Detector:
            # Panel:
            # pixel_size:{0.075,0.075}
            # image_size: {2070,2167}
            # trusted_range: {0,1e+07}
            # thickness: 0
            # material:
            # mu: 0
            # fast_axis: {-0.999993,0.00338011,-0.00156153}
            # slow_axis: {0.00213163,0.863573,0.504219}
            # origin: {75.1912,14.1117,-124.34}

            detector = self._detector_factory.complex(
                "PAD",
                origin=(75.1912, 14.1117, -124.34),
                fast=(-0.999993, 0.00338011, -0.00156153),
                slow=(0.00213163, 0.863573, 0.504219),
                pixel=(0.075, 0.075),
                size=(2070, 2167),
                trusted_range=(0, max_trusted_value),
            )

        else:
            raise RuntimeError("Don't understand image")

        return detector

    def _beam(self):
        # Beam:
        # wavelength: 1.1
        # sample to source direction : {-0.00573979,-0,0.999984}
        # divergence: 0
        # sigma divergence: 0
        # polarization normal: {0,1,0}
        # polarization fraction: 0.999

        return self._beam_factory.complex(
            sample_to_source=(-0.00573979, -0, 0.999984),
            polarization_fraction=0.999,
            polarization_plane_normal=(0, 1, 0),
            wavelength=1.1,
        )

    def _scan(self):
        exposure_time = 1  # XXX
        osc_start = float(self._cif_header_dictionary["Start_angle"].split()[0])
        osc_range = 0.1
        timestamp = 1  # XXX

        return self._scan_factory.single_file(
            self._image_file, exposure_time, osc_start, osc_range, timestamp
        )

    def detectorbase_start(self):

        self.detectorbase = PilatusImage(self._image_file)
        self.detectorbase.readHeader()

        self.detectorbase.parameters["SIZE1"] = 2167
        self.detectorbase.parameters["SIZE2"] = 2070

    def get_vendortype(self):
        return gv(self.get_detector())


if __name__ == "__main__":
    for arg in sys.argv[1:]:
        print(FormatCBFMiniEigerPhotonFactory.understand(arg))



from __future__ import annotations

from typing import Any, Tuple, Union

from scitbx.array_family import flex

ImageData = Union[flex.int, flex.double, flex.float]

class ExternalLookup:
    def __init__(self, *args, **kwargs) -> None: ...
    def __reduce__(self) -> Any: ...
    @property
    def dx(self) -> Any: ...
    @property
    def dy(self) -> Any: ...
    @property
    def gain(self) -> Any: ...
    @property
    def mask(self) -> Any: ...
    @property
    def pedestal(self) -> Any: ...

class ExternalLookupItemBool:

    data: Any
    filename: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def __reduce__(self) -> Any: ...

class ExternalLookupItemDouble:

    data: Any
    filename: Any
    def __init__(self, *args, **kwargs) -> None: ...
    def __reduce__(self) -> Any: ...

class ImageGrid(ImageSet):
    def __init__(self, *args, **kwargs) -> None: ...
    def from_imageset(self, *args, **kwargs) -> Any: ...
    def get_grid_size(self) -> Any: ...
    def __getinitargs__(self) -> Any: ...
    def __reduce__(self) -> Any: ...

class ImageSequence(ImageSet):
    def __init__(self, *args, **kwargs) -> None: ...
    def complete_set(self) -> Any: ...
    def get_array_range(self) -> Any: ...
    def get_beam(self) -> Any: ...
    def get_detector(self) -> Any: ...
    def get_goniometer(self) -> Any: ...
    def get_scan(self) -> Any: ...
    def partial_set(self, *args, **kwargs) -> Any: ...
    def set_beam(self, boost) -> Any: ...
    def set_detector(self, boost) -> Any: ...
    def set_goniometer(self, boost) -> Any: ...
    def set_scan(self, boost) -> Any: ...
    def update_detector_px_mm_data(self) -> Any: ...
    def __getinitargs__(self) -> Any: ...
    def __reduce__(self) -> Any: ...

class ImageSet:
    def __init__(boost, dxtbx) -> None: ...
    def as_imageset(self) -> Any: ...
    def clear_cache(self) -> Any: ...
    def complete_set(self) -> Any: ...
    def data(self) -> Any: ...
    def get_beam(self) -> Any: ...
    def get_corrected_data(self, int) -> Any: ...
    def get_detector(self) -> Any: ...
    def get_gain(self, int) -> Any: ...
    def get_goniometer(self) -> Any: ...
    def get_image_identifier(self, int) -> Any: ...
    def get_mask(self, int) -> Any: ...
    def get_path(self, int) -> Any: ...
    def get_pedestal(self, int) -> Any: ...
    def get_raw_data(self, index: int) -> Tuple[ImageData]: ...
    def get_scan(self) -> Any: ...
    def has_dynamic_mask(self) -> Any: ...
    def indices(self) -> Any: ...
    def is_marked_for_rejection(self, int) -> Any: ...
    def mark_for_rejection(self, int, bool) -> Any: ...
    def partial_set(self, *args, **kwargs) -> Any: ...
    def set_beam(self, boost) -> Any: ...
    def set_detector(self, boost) -> Any: ...
    def set_goniometer(self, boost) -> Any: ...
    def set_scan(self, boost) -> Any: ...
    def size(self) -> Any: ...
    def update_detector_px_mm_data(self) -> Any: ...
    def __eq__(self, other) -> Any: ...
    def __getinitargs__(self) -> Any: ...
    def __len__(self) -> Any: ...
    def __ne__(self, other) -> Any: ...
    def __reduce__(self) -> Any: ...
    @property
    def external_lookup(self) -> Any: ...

class ImageSetData:
    def __init__(self, *args, **kwargs) -> None: ...
    def get_beam(self, int) -> Any: ...
    def get_data(self, int) -> Any: ...
    def get_detector(self, int) -> Any: ...
    def get_format_class(self) -> Any: ...
    def get_goniometer(self, int) -> Any: ...
    def get_image_identifier(self, int) -> Any: ...
    def get_master_path(self) -> Any: ...
    def get_params(self) -> Any: ...
    def get_path(self, int) -> Any: ...
    def get_scan(self, int) -> Any: ...
    def get_template(self) -> Any: ...
    def get_vendor(self) -> Any: ...
    def has_single_file_reader(self) -> Any: ...
    def is_marked_for_rejection(self, int) -> Any: ...
    def mark_for_rejection(self, int, bool) -> Any: ...
    def masker(self) -> Any: ...
    def reader(self) -> Any: ...
    def set_beam(self, boost, int) -> Any: ...
    def set_detector(self, boost, int) -> Any: ...
    def set_format_class(self, boost) -> Any: ...
    def set_goniometer(self, boost, int) -> Any: ...
    def set_params(self, boost) -> Any: ...
    def set_scan(self, boost, int) -> Any: ...
    def set_template(self, *args, **kwargs) -> Any: ...
    def set_vendor(self, *args, **kwargs) -> Any: ...
    def __getinitargs__(self) -> Any: ...
    def __getstate__(self) -> Any: ...
    def __reduce__(self) -> Any: ...
    def __setstate__(self, boost) -> Any: ...
    @property
    def external_lookup(self) -> Any: ...



from __future__ import annotations

import argparse

import dxtbx.util
from dxtbx import load


def saturation(image_file):
    i = load(image_file)
    d = i.get_detector()
    raw_data = i.get_raw_data()
    if not isinstance(raw_data, tuple):
        raw_data = (raw_data,)
    if i.get_scan() is None:
        return (
            0,
            max(
                max(raw_data[pid]) / detector.get_trusted_range()[1]
                for pid, detector in enumerate(d)
            ),
        )
    else:
        return (
            i.get_scan().get_image_range()[0],
            max(
                max(raw_data[pid]) / detector.get_trusted_range()[1]
                for pid, detector in enumerate(d)
            ),
        )


def run(args=None):
    dxtbx.util.encode_output_as_utf8()
    parser = argparse.ArgumentParser()
    parser.add_argument("images", metavar="IMAGE", help="Image files", nargs="+")
    options = parser.parse_args(args)

    for image_file in options.images:
        i, s = saturation(image_file)
        print(f"{i:6d} {s:.6f}")


if __name__ == "__main__":
    run()



from __future__ import annotations

import pytest

from dxtbx.model.experiment_list import ExperimentListFactory

"""
Test deserializing an experiment list that has single file indices while using check_format = True or False
"""


def test_split_single_image_imageset(dials_data, tmpdir):
    pytest.importorskip("h5py")
    sacla_file = (
        dials_data("image_examples", pathlib=True) / "SACLA-MPCCD-run266702-0-subset.h5"
    )
    expts = ExperimentListFactory.from_filenames([sacla_file])
    assert len(expts) == 4
    subset = expts[2:3]
    assert len(subset) == 1
    assert list(subset[0].imageset.indices()) == [2]

    dumped_filename = tmpdir / "split.expt"
    subset.as_file(dumped_filename)

    subset = ExperimentListFactory.from_json_file(dumped_filename, check_format=True)
    assert len(subset) == 1
    assert list(subset[0].imageset.indices()) == [2]

    subset = ExperimentListFactory.from_json_file(dumped_filename, check_format=False)
    assert len(subset) == 1
    assert list(subset[0].imageset.indices()) == [2]
